/*
 * Minimum Comparison Sorting --- Experiments
 *
 * main.c
 * Main module
 */

/* Copyright 2002-2005, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "main.h"

static void main_init(void);
static void main_clean(void);
static int  print_step(BOOL);
static int  extract_posets(long, long);
static int  compare_posets(long, long);
static int  get_component_cardinality(const POSET*, int);

extern int   file[];
extern POSET read_buffer[];

#ifdef _USE_POF2_TABLE_
extern SET           pof2_tbl[];
#endif
#ifdef _USE_MIN_ELEMENT_TABLE_
extern MIN_EL_TBL_EL min_element_tbl[];
#endif
#ifdef _USE_CARDINALITY_TABLE_
extern CARD_TBL_EL   cardinality_tbl[];
#endif

/*
 * Global data initialization and deinitialization
 */

static void main_init(void) {
  sort_module_init();
  cpcp_module_init();
}

static void main_clean(void) {
  int i;

  for (i = 0; i < 3; i++)
    if (file[i] >= 0)
      close(file[i]);
  sort_module_clean();
  cpcp_module_clean();
}

/*
 * Program administration
 */

void fatal_error(const char *fmt, ...) {
  va_list fmt_args;

  fprintf(stderr, "ERROR: ");
  va_start(fmt_args, fmt);
  vfprintf(stderr, fmt, fmt_args);
  va_end(fmt_args);
  if (errno) {
    fprintf(stderr, ", errno=%d", errno);
    perror(" ");
  }
  else
    fprintf(stderr, "\n");
  main_clean();
  exit(2);
}

int main(int argc, char *argv[]) {
  int result = 0;

  if (argc < 2) {
    printf("usage: %s action [parameters]\ncompiled at "__DATE__
           " "__TIME__"\n", argv[0]);
    #ifdef _NRCT_
    printf("non-recursive congruence testing\n");
    #else
    printf("recursive congruence testing\n");
    #endif
    #ifdef _INT_INDEX_
    printf("int indices\n");
    #else
    printf("long indices\n");
    #endif
    #if defined _USE_MIN_ELEMENT_TABLE_
      printf("min element table\n");
    #elif defined _USE_BUILT_IN_MIN_ELEMENT_
      printf("built-in min element\n");
    #endif
    #if defined _USE_BUILT_IN_MAX_ELEMENT_
      printf("built-in max element\n");
    #endif
    #if THREADS_NUMBER > 1
    printf("%d threads\n", THREADS_NUMBER);
    #ifdef _RW_ALG_
    printf("readers/writers algorithm\n");
    #else
    printf("simple mutex\n");
    #endif
    #else
    printf("%d thread\n", THREADS_NUMBER);
    #endif
    #if defined _WIDTH_2_
      printf("M = %d   ", M);
    #endif
    printf("N = %d   sizeof(POSET) = %d\n"
           "MAX_STACK = %ld   HASH_TBL_SIZE = %ld   SORT_TBL_SIZE = %ld\n"
           "UPDATE_INTERVAL = %d   READ_INTERVAL = %d\n\n",
           N, (int)sizeof(POSET),
           (long)MAX_STACK, (long)HASH_TBL_SIZE, (long)SORT_TBL_SIZE,
           (int)UPDATE_INTERVAL, (int)READ_INTERVAL);
    return 3;
  }
  switch (atoi(argv[1])) {

    case 0: /* Create initial file containing the $S_c$ set. */
      if (argc < 4) {
        printf("usage: %s 0 out_file c_max [c]\n\n", argv[0]);
        return 3;
      }
      file[OUTF] = open(argv[2], O_RDWR | O_CREAT | O_TRUNC | O_BINARY, FATR);
      if (file[OUTF] < 0)
        fatal_error("can't create file %s", argv[2]);
      main_init();
      result = first_step(atoi(argv[3]), argc == 5 ? atoi(argv[4]) : 1);
      break;

    case 1: /* Process file containing $S_{c-1}$ set.
               Produce file with $S_c$ set. */
      if (argc < 4) {
        printf("usage: %s 1 in_file out_file\n\n", argv[0]);
        return 3;
      }
      file[INF] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[INF] < 0)
        fatal_error("can't open file %s", argv[2]);
      file[OUTF] = open(argv[3], O_RDWR | O_CREAT | O_EXCL | O_BINARY, FATR);
      if (file[OUTF] < 0 && errno == EEXIST) {
        file[OUTF] = open(argv[3], O_RDWR | O_BINARY);
        if (file[OUTF] < 0) {
          fatal_error("can't open file %s", argv[3]);
        }
        else {
          errno = 0;
          main_init();
          result = one_step(TRUE);
        }
      }
      else if (file[OUTF] < 0) {
        fatal_error("can't create file %s", argv[3]);
      }
      else {
        main_init();
        result = one_step(FALSE);
      }
      break;

    case 2: /* Create initial file with one poset containing
               two chains: m-element and (N - m)-element. */
      #ifdef _WIDTH_2_
      if (argc < 4) {
        printf("usage: %s 2 out_file c_max\n\n", argv[0]);
        return 3;
      }
      file[OUTF] = open(argv[2], O_RDWR | O_CREAT | O_TRUNC | O_BINARY, FATR);
      if (file[OUTF] < 0)
        fatal_error("can't create file %s", argv[2]);
      main_init();
      result = two_chains(M, atoi(argv[3]));
      #else
      if (argc < 5) {
        printf("usage: %s 2 out_file m c_max\n\n", argv[0]);
        return 3;
      }
      file[OUTF] = open(argv[2], O_RDWR | O_CREAT | O_TRUNC | O_BINARY, FATR);
      if (file[OUTF] < 0)
        fatal_error("can't create file %s", argv[2]);
      main_init();
      result = two_chains(atoi(argv[3]), atoi(argv[4]));
      #endif
      break;

    case 3: /* Find all sortable posets in input file $S_c$.
               Sortable posets are stored in output file $S^*_c$. */
      if (argc < 4) {
        printf("usage: %s 3 in_file out_file\n\n", argv[0]);
        return 3;
      }
      file[INF] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[INF] < 0)
        fatal_error("can't open file %s", argv[2]);
      file[OUTF] = open(argv[3], O_RDWR | O_CREAT | O_EXCL | O_BINARY, FATR);
      if (file[OUTF] < 0 && errno == EEXIST) {
        file[OUTF] = open(argv[3], O_RDWR | O_BINARY);
        if (file[OUTF] < 0) {
          fatal_error("can't open file %s", argv[3]);
        }
        else {
          errno = 0;
          main_init();
          result = alg_search(TRUE);
        }
      }
      else if (file[OUTF] < 0) {
        fatal_error("can't create file %s", argv[3]);
      }
      else {
        main_init();
        result = alg_search(FALSE);
      }
      break;

    case 4: /* Find all sortable posets in input file $S_c$.
               Sortable posets are stored in output file $S^*_c$.
               Sortable file $S^*_{c-1}$ contains sortable posets found
               in previous step. */
      if (argc < 5) {
        printf("usage: %s 4 in_file out_file sortable_file [match_type]\n\n",
               argv[0]);
        return 3;
      }
      file[INF] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[INF] < 0)
        fatal_error("can't open file %s", argv[2]);
      file[SORTF] = open(argv[4], O_RDONLY | O_BINARY);
      if (file[SORTF] < 0)
        fatal_error("can't open file %s", argv[4]);
      file[OUTF] = open(argv[3], O_RDWR | O_CREAT | O_EXCL | O_BINARY, FATR);
      if (file[OUTF] < 0 && errno == EEXIST) {
        file[OUTF] = open(argv[3], O_RDWR | O_BINARY);
        if (file[OUTF] < 0) {
          fatal_error("can't open file %s", argv[3]);
        }
        else {
          errno = 0;
          main_init();
          if (argc == 6)
            fatal_error("alg_search_3, recovery not implemented");
          else
            result = alg_search_2(TRUE);
        }
      }
      else if (file[OUTF] < 0) {
        fatal_error("can't create file %s", argv[3]);
      }
      else {
        main_init();
        if (argc == 6)
          result = alg_search_3(atoi(argv[5]));
        else
          result = alg_search_2(FALSE);
      }
      break;

    case 5: /* Display overview of file with $S$ set. */
      if (argc < 3) {
        printf("usage: %s 5 in_file\n\n", argv[0]);
        return 3;
      }
      file[INF] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[INF] < 0)
        fatal_error("can't open file %s", argv[2]);
      main_init();
      result = print_step(FALSE);
      break;

    case 6: /* Display contents of file with $S$ set. */
      if (argc < 3) {
        printf("usage: %s 6 in_file\n\n", argv[0]);
        return 3;
      }
      file[INF] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[INF] < 0)
        fatal_error("can't open file %s", argv[2]);
      main_init();
      result = print_step(TRUE);
      break;

    case 7: /* Create initial file with one poset. */
      if (argc < 4) {
        printf("usage: %s 7 out_file poset_type c_max\n\n", argv[0]);
        return 3;
      }
      file[OUTF] = open(argv[2], O_RDWR | O_CREAT | O_TRUNC | O_BINARY, FATR);
      if (file[OUTF] < 0)
        fatal_error("can't create file %s", argv[2]);
      main_init();
      result = special_poset(atoi(argv[3]), atoi(argv[4]));
      break;

    case 8: { /* Create initial file from text description. */
      int c_max;

      c_max = -1;
      switch (argc) {
        case 5:
          c_max = atoi(argv[4]);
        case 4:
          file[OUTF] = open(argv[3], O_RDWR | O_CREAT | O_TRUNC | O_BINARY,
                            FATR);
          if (file[OUTF] < 0)
            fatal_error("can't create file %s", argv[3]);
        case 3:
          main_init();
          result = generate_from_description(argv[2], c_max);
          break;
        default:
          printf("usage: %s 8 in_file [out_file [c_max]]\n\n", argv[0]);
          return 3;
      }
      break;
    }

    case 9: { /* Extract posets from in_file to out_file. */
      #define UPPER_LIMIT MAX_STACK
      #define LOWER_LIMIT (3 * (MAX_STACK / 4))
      long from, morf, to, s, p;

      if (argc == 6) {
        from = atol(argv[4]);
        to   = atol(argv[5]);
      }
      else if (argc == 8) {
        from  = atol(argv[4]);
        morf  = atol(argv[5]);
        s     = atol(argv[6]);
        p     = atol(argv[7]);
        if (p > UPPER_LIMIT)
          s = (s + (long)((double)s * (double)UPPER_LIMIT / (double)p)) / 2;
        else if (p < LOWER_LIMIT && p > 0)
          s = (s + (long)((double)s * (double)LOWER_LIMIT / (double)p)) / 2;
        to = from + s - 1;
        if (to > morf)
          to = morf;
      }
      else {
        printf("usage: %s 9 in_file out_file from_poset to_poset "
               "[prev_step prev_max]\n\n", argv[0]);
        return 3;
      }
      file[INF] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[INF] < 0)
        fatal_error("can't open file %s", argv[2]);
      file[OUTF] = open(argv[3], O_RDWR | O_CREAT | O_EXCL | O_BINARY, FATR);
      if (file[OUTF] < 0 && errno == EEXIST) {
        file[OUTF] = open(argv[3], O_RDONLY | O_BINARY);
        if (file[OUTF] < 0) {
          fatal_error("can't open file %s", argv[3]);
        }
        else {
          errno = 0;
          result = compare_posets(from, to);
        }
      }
      else if (file[OUTF] < 0) {
        fatal_error("can't create file %s", argv[3]);
      }
      else {
        result = extract_posets(from, to);
      }
      break;
    }

    case 10: /* Append sortable posets from in_file to out_file. */
      if (argc < 4) {
        printf("usage: %s 10 in_file out_file\n\n", argv[0]);
        return 3;
      }
      file[INF] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[INF] < 0)
        fatal_error("can't open file %s", argv[2]);
      file[OUTF] = open(argv[3], O_RDWR | O_BINARY);
      if (file[OUTF] < 0)
        fatal_error("can't open file %s", argv[3]);
      main_init();
      result = join_posets();
      break;

    #if THREADS_NUMBER == 1

    case 11: /* Store in out_file all posets from in_file_2
                not congruent to posets from in_file_1. */
      if (argc < 5) {
        printf("usage: %s 11 in_file_1 in_file_2 out_file\n\n", argv[0]);
        return 3;
      }
      file[IN1F] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[IN1F] < 0)
        fatal_error("can't open file %s", argv[2]);
      file[IN2F] = open(argv[3], O_RDONLY | O_BINARY);
      if (file[IN2F] < 0)
        fatal_error("can't open file %s", argv[3]);
      file[OUTF] = open(argv[4], O_RDWR  | O_CREAT | O_EXCL | O_BINARY, FATR);
      if (file[OUTF] < 0)
        fatal_error("can't create file %s", argv[4]);
      main_init();
      result = remove_congruent_posets();
      break;

    case 12: /* Select all posets with t <= 2^c_max. */
      if (argc < 5) {
        printf("usage: %s 12 in_file out_file c_max\n\n", argv[0]);
        return 3;
      }
      file[INF] = open(argv[2], O_RDONLY | O_BINARY);
      if (file[INF] < 0)
        fatal_error("can't open file %s", argv[2]);
      file[OUTF] = open(argv[3], O_RDWR | O_CREAT | O_EXCL | O_BINARY, FATR);
      if (file[OUTF] < 0)
        fatal_error("can't create file %s", argv[3]);
      main_init();
      result = select_posets(atoi(argv[4]));
      break;

    case 14: /* Test the Gold Partition Conjecture. */
      /*
         test_type:  1 - GPC
                     2 - the worst balanced poset
                     4 - B-set
         chech_type: 0 - not a linear sum
                     2 - height 2
                     3 - not a linear sum, width at most 3
                     4 - not a linear sum, ladders with broken rungs
      */
      if (argc < 4) {
        printf("usage: %s 14 test_type check_type\n\n", argv[0]);
        return 3;
      }
      main_init();
      result = gpc_iterative_test(atoi(argv[2]), atoi(argv[3]));
      break;

    #endif /* THREADS_NUMBER == 1 */

    default:
      printf("unknown action: %s\n", argv[1]);
      return 3;
  }
  main_clean();
  return result;
}

/*
 * Auxiliary routines
 */

static int print_step(BOOL print_posets) {
  FILE_HEADER fh;
  POSET       *p_g;
  CPC         t_min, t_max;
  long        i, v[N + 1], c[N + 1], v_sum, c_sum;

  if (read(file[INF], &fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("print_header, read file header failed");
  if (fh.version != FILE_VERSION)
    fatal_error("print_step, missing input file version %d", fh.version);

  printf("file version %d  step %d  input posets %ld  output posets %ld\n"
         "cut off " CPC_PRINT_FORMAT "  computation time %ld\n",
         fh.version, fh.step, fh.analysed_count, fh.posets_count,
         fh.cut_off, fh.time);

  for (i = 0; i <= N; i++)
    v[i] = c[i] = 0;
  t_min = t_max = 0;

  reset_read_buffer(INF);
  for (i = 0; i < fh.posets_count; i++) {
    p_g = get_next_poset(NULL);
    if (print_posets) {
      printf("*%ld* ", i);
      poset_print(p_g);
    }
    v[p_g->n]++;
    if (p_g->n == get_component_cardinality(p_g, 0))
      c[p_g->n]++;
    if (p_g->t < t_min || t_min == 0)
      t_min = p_g->t;
    if (p_g->t > t_max)
      t_max = p_g->t;
  }

  printf("-------------------------\n"
         "elements posets connected\n");
  v_sum = c_sum = 0;
  for (i = 0; i <= N; i++) {
    if (v[i]) {
      v_sum += v[i];
      c_sum += c[i];
      printf("%3ld %10ld %10ld\n", i, v[i], c[i]);
    }
  }
  printf("-------------------------\n"
         "sum %10ld %10ld\n"
         "-------------------------\n"
         "min e(P)=" CPC_PRINT_FORMAT "   max e(P)=" CPC_PRINT_FORMAT "\n",
         v_sum, c_sum, t_min, t_max);

  return fh.posets_count == 0;
}

static int extract_posets(long from, long to) {
  FILE_HEADER fh;
  long        i;

  if (read(file[INF], &fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("extract_posets, read file header failed");
  if (fh.version != FILE_VERSION)
    fatal_error("extract_posets, missing input file version %d",
                fh.version);
  if (lseek(file[INF], sizeof(FILE_HEADER) + from*sizeof(POSET), SEEK_SET) < 0)
    fatal_error("extract_posets, lseek failed");
  if (from < 0 || to < 0 || to < from || from >= fh.posets_count)
    fatal_error("extract_posets, missing range, from %ld to %ld", from, to);
  if (to > fh.posets_count - 1)
    to = fh.posets_count - 1;
  fh.analysed_count = fh.posets_count;
  fh.posets_count = to - from + 1;
  fh.time = 0;
  if (write(file[OUTF], &fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("extract_posets, write output file header failed");
  for (i = from; i <= to - READ_INTERVAL + 1; i += READ_INTERVAL) {
    if (read(file[INF], read_buffer, READ_INTERVAL * sizeof(POSET)) !=
        READ_INTERVAL * sizeof(POSET))
      fatal_error("extract_posets, read(1) failed");
    if (write(file[OUTF], read_buffer, READ_INTERVAL * sizeof(POSET)) !=
        READ_INTERVAL * sizeof(POSET))
      fatal_error("extract_posets, write(1) failed");
  }
  i = to - i + 1;
  if (i > 0) {
    i *= sizeof(POSET);
    if (read(file[INF], read_buffer, i) != i)
      fatal_error("extract_posets, read(2) failed");
    if (write(file[OUTF], read_buffer, i) != i)
      fatal_error("extract_posets, write(2) failed");
  }

  printf("step %d  input posets %ld  output posets %ld (%ld-%ld)\n",
         fh.step, fh.analysed_count, fh.posets_count, from, to);
  return fh.posets_count == 0;
}

static int compare_posets(long from, long to) {
  POSET       out_buffer[READ_INTERVAL];
  FILE_HEADER in_fh, out_fh;
  long        i, j;

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("compare_posets: can't read input file header");
  if (read(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("compare_posets: can't read output file header");
  if (from < 0 || to < 0 || to < from || from >= in_fh.posets_count)
    fatal_error("compare_posets: missing range: from %ld to %ld", from, to);
  if (in_fh.version != FILE_VERSION ||
      out_fh.version != FILE_VERSION ||
      in_fh.step != out_fh.step ||
      in_fh.cut_off != out_fh.cut_off)
    fatal_error("compare_posets: missing headers");
  if (to > in_fh.posets_count - 1)
    to = in_fh.posets_count - 1;
  if (out_fh.posets_count != to - from + 1)
    fatal_error("compare_posets: missing number of posets");
  if (lseek(file[INF], sizeof(FILE_HEADER) + from*sizeof(POSET), SEEK_SET) < 0)
    fatal_error("compare_posets: call to lseek failed");
  for (i = from; i <= to - READ_INTERVAL + 1; i += READ_INTERVAL) {
    if (read(file[INF], read_buffer, READ_INTERVAL * sizeof(POSET)) !=
        READ_INTERVAL * sizeof(POSET))
      fatal_error("compare_posets: can't read poset from input fille (1)");
    if (read(file[OUTF], out_buffer, READ_INTERVAL * sizeof(POSET)) !=
        READ_INTERVAL * sizeof(POSET))
      fatal_error("compare_posets: can't read poset from output file (1)");
    if (memcmp(read_buffer, out_buffer, READ_INTERVAL * sizeof(POSET)))
      fatal_error("compare_posets: different poset found %ld (1)", i);
  }
  j = to - i + 1;
  if (j > 0) {
    j *= sizeof(POSET);
    if (read(file[INF], read_buffer, j) != j)
      fatal_error("compare_posets: can't read poset from input fille (2)");
    if (read(file[OUTF], out_buffer, j) != j)
      fatal_error("compare_posets: can't read poset from output file (2)");
    if (memcmp(read_buffer, out_buffer, j))
      fatal_error("compare_posets: different poset found %ld (2)", i);
  }

  printf("step %d  input posets %ld  output posets %ld (%ld-%ld)\n",
         out_fh.step, out_fh.analysed_count, out_fh.posets_count, from, to);
  return out_fh.posets_count == 0;
}

static int get_component_cardinality(const POSET *g, int v) {
  SET prev_component, component, new_elements;

  prev_component = pof2(v);
  component = prev_component | g->r[0][v] | g->r[1][v];
  while(component != prev_component) {
    new_elements = component & ~prev_component;
    prev_component = component;
    while (new_elements) {
      v = min_element(new_elements);
      new_elements ^= pof2(v);
      component |= g->r[0][v] | g->r[1][v];
    }
  }
  return cardinality(component);
}

void poset_print(const POSET *g) {
  int i, j, k;
  SET V;
  static const char description[2][6] = {"above", "below"};

  printf("n=%d e(P)=" CPC_PRINT_FORMAT "\n", g->n, g->t);
  for (k = 0; k < 2; k++) {
    printf("%s\n", description[k]);
    for (i = 0; i < g->n; i++) {
      V = g->r[k][i];
      if (V) {
        printf("%d:", i);
        do {
          j = min_element(V);
          V ^= pof2(j);
          printf(" %d", j);
        } while(V);
        printf("\n");
      }
    }
  }
}
