/*
 * Minimum Comparison Sorting --- Experiments
 *
 * main.h
 * General definitions
 */

/* Copyright 2002-2005, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#ifndef _MAIN_H_
#define _MAIN_H_

/*
 * Compilation control constants (currently defined in makefile)
 *
 * Architecture definition:
 *   __WIN32__             -- Win32 on 32-bit x86 machine
 *   _IA32_                -- Linux on 32-bit x86 machine
 *   _X86_64_              -- Linux on 64-bit x86 machine
 *   _ILP32_               -- ILP32 machine with 64-bit long long type
 *   _LP64_                -- LP64 machine with 64-bit long type
 *
 * The experiment has to be defined:
 *   N                     -- the number of elements
 *   M                     -- width-2 posets, M and N - M element chains
 *
 * Other possible definition:
 *   _PRINT_PROGRESS_      -- progress information is displayed during
 *                            experiment
 *   _NRCT_                -- non-recursive congruence testing
 *   _INT_INDEX_           -- int indices in hash table, use carefully if
 *                            MAX_STACK * sizeof(POSET) > INT_MAX
 *
 * Multithreading:
 *   THREADS_NUMBER        -- the number of threads
 *   _RW_ALG_              -- access to hash table using readers/writers
 *                            algorithm insted of a simple mutex,
 *                            for only two threads mutex seems to be faster
 */

#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __WIN32__
#include <io.h>
#else
#define O_BINARY 0
#include <unistd.h>
#endif

#ifndef THREADS_NUMBER
  #define THREADS_NUMBER 1
#endif

/* Intermediate results are stored to disk
   after processing each UPDATE_INTERVAL posets. */
#define UPDATE_INTERVAL 9000

/* Posets are read from disk in READ_INTERVAL blocks. */
#define READ_INTERVAL 9000

/*
 * Experiment definition
 */

#ifdef M
  #define _WIDTH_2_
#endif

#ifndef _CONFIG_
  #include "config.h"
#endif

/*
 * Common types, functions and macros
 */

typedef int BOOL;
#define TRUE  1
#define FALSE 0

#if defined(__WIN32__) && !defined(_IA32_)
  #define _IA32_
#endif

#if defined(_IA32_) && N <= 12
  typedef long CPC;
  #define CPC_PRINT_FORMAT "%ld"
#elif defined(_IA32_) && N <= 18
  typedef double CPC;
  #define CPC_PRINT_FORMAT "%-.0f"
#elif defined(_IA32_)
  typedef long double CPC;
  #define CPC_PRINT_FORMAT "%-.0Lf"
#elif defined(_X86_86_) && N <= 12
  typedef int CPC;
  #define CPC_PRINT_FORMAT "%d"
#elif defined(_X86_64_)
  typedef long CPC;
  #define CPC_PRINT_FORMAT "%ld"
#elif defined(_ILP32_) && N <= 12
  typedef int CPC;
  #define CPC_PRINT_FORMAT "%d"
#elif defined(_ILP32_) && N <= 20
  typedef long long CPC;
  #define CPC_PRINT_FORMAT "%Ld"
#elif defined(_LP64_) && N <= 12
  typedef int CPC;
  #define CPC_PRINT_FORMAT "%d"
#elif defined(_LP64_) && N <= 20
  typedef long CPC;
  #define CPC_PRINT_FORMAT "%ld"
#else
  #error The number of linear extensions cannot be counted for this N.
#endif

#if N <= 16
  typedef unsigned short SET;
#elif N <= 32
  typedef unsigned int SET;
#elif N <= 64 && (defined(_X86_64_) || defined(_LP64_))
  typedef unsigned long SET;
#elif N <= 64 && (defined(_IA32_) || defined(_ILP32_))
  typedef unsigned long long SET;
#else
  #error N is too large.
#endif
#define EMPTY_SET 0

#ifdef _INT_INDEX_
  typedef int INDEX;
#else
  typedef long INDEX;
#endif

#if defined(_IA32_) || defined(_X86_64_)
  #define _USE_POF2_TABLE_
  #if defined(__GNUC__) && !defined(__STRICT_ANSI__)
    #define _USE_BUILT_IN_MIN_ELEMENT_
    #define _USE_BUILT_IN_MAX_ELEMENT_
  #else
    #define _USE_MIN_ELEMENT_TABLE_
    typedef char MIN_EL_TBL_EL;
  #endif
  #define _USE_CARDINALITY_TABLE_
  typedef char CARD_TBL_EL;
#else
  #define _USE_MIN_ELEMENT_TABLE_
  typedef int MIN_EL_TBL_EL;
  #define _USE_CARDINALITY_TABLE_
  typedef int CARD_TBL_EL;
#endif

#ifdef _USE_POF2_TABLE_
  /* To look in a table is faster than shift operation! */
  #define pof2(x) pof2_tbl[x]
  #define in(s, x) (((s) & pof2_tbl[x]) != EMPTY_SET)
#else
  #define pof2(x) (1 << (x))
  #define in(s, x) (((s) & (1 << (x))) != EMPTY_SET)
#endif

#if defined(_USE_MIN_ELEMENT_TABLE_)
  #if N <= 24
    #define MET_SIZE       (1 << N)
    #define min_element(x) (int)min_element_tbl[x]
  #elif N <= 48
    #define MET_HALF_N     ((N + 1) / 2)
    #define MET_SIZE       (1 << MET_HALF_N)
    #define MET_HALF_SET   (FULL_SET >> (N / 2))
    #define min_element(x) ((x) & MET_HALF_SET ?                             \
                            min_element_tbl[(x) & MET_HALF_SET] :            \
                            min_element_tbl[(x) >> MET_HALF_N] + MET_HALF_N)
  #else
    #error Minimum element table too big.
  #endif
#elif defined(_USE_BUILT_IN_MIN_ELEMENT_)
  #if defined(_IA32_) && N <= 32
    #define min_element(set) ({                                              \
      register int element;                                                  \
      asm ("bsfl %1, %0" : "=r"(element) : "g"((unsigned int)(set)) : "cc"); \
      element;                                                               \
    })
  #elif defined(_X86_64_) && N <= 16
    #define min_element(set) ({                                     \
      register short element;                                       \
      asm ("bsfw %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #elif defined(_X86_64_) && N <= 32
    #define min_element(set) ({                                     \
      register int element;                                         \
      asm ("bsfl %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #elif defined(_X86_64_) && N <= 64
    #define min_element(set) ({                                     \
      register long element;                                        \
      asm ("bsfq %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #else
    #error There is no built-in min_element function.
  #endif
#else
  int min_element(SET);
#endif

#if defined(_USE_BUILT_IN_MAX_ELEMENT_)
  #if defined(_IA32_) && N <= 32
    #define max_element(set) ({                                              \
      register int element;                                                  \
      asm ("bsrl %1, %0" : "=r"(element) : "g"((unsigned int)(set)) : "cc"); \
      element;                                                               \
    })
  #elif defined(_X86_64_) && N <= 16
    #define max_element(set) ({                                     \
      register short element;                                       \
      asm ("bsrw %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #elif defined(_X86_64_) && N <= 32
    #define max_element(set) ({                                     \
      register int element;                                         \
      asm ("bsrl %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #elif defined(_X86_64_) && N <= 64
    #define max_element(set) ({                                     \
      register long element;                                        \
      asm ("bsrq %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #else
    #error There is no built-in max_element function.
  #endif
#else
  int max_element(SET);
#endif

#if defined(_USE_CARDINALITY_TABLE_)
  #if N <= 24
    #define CAT_SIZE       (1 << N)
    #define cardinality(x) (int)cardinality_tbl[x]
  #elif N <= 48
    #define CAT_HALF_N     ((N + 1) / 2)
    #define CAT_SIZE       (1 << CAT_HALF_N)
    #define CAT_HALF_SET   (FULL_SET >> (N / 2))
    #define cardinality(x) (cardinality_tbl[(x) & CAT_HALF_SET] + \
                            cardinality_tbl[(x) >> CAT_HALF_N])
  #else
    #error Cardinality table too big.
  #endif
#elif defined(_USE_BUILT_IN_CARDINALITY_)
  #error There is not built-in cardinality function.
#else
  int cardinality(SET);
#endif

#if N <= 33
  /* 32-bit integer is sufficient and faster. */
  typedef int BINOM;
#else
  typedef CPC BINOM;
#endif

#ifdef M
  #define FILE_VERSION 100*M + N
#else
  #define FILE_VERSION N
#endif

typedef struct {
  CPC   t;
  INDEX hash;
  int   n;
  SET   r[2][N];
} POSET;

#define poset_empty(g) memset(g, 0, sizeof(POSET))

typedef struct {
  int    version;
  int    step;
  long   analysed_count;
  long   posets_count;
  time_t time;
  CPC    cut_off;
} FILE_HEADER;

#define file_header_empty(f) memset(f, 0, sizeof(FILE_HEADER))
#define FATR 0644 /* atributes for a new file */
#define INF   0
#define OUTF  1
#define SORTF 2
#define IN1F  INF
#define IN2F  SORTF

void   sort_module_init(void);
void   sort_module_clean(void);
void   cpcp_module_init(void);
void   cpcp_module_clean(void);
void   fatal_error(const char*, ...);
void   reset_read_buffer(int);
POSET* get_next_poset(POSET*);
int    one_step(BOOL);
int    join_posets(void);
int    remove_congruent_posets(void);
int    alg_search(BOOL);
int    alg_search_2(BOOL);
int    alg_search_3(int);
int    first_step(int, int);
int    two_chains(int, int);
int    special_poset(int, int);
int    generate_from_description(const char*, int);
int    select_posets(int);
int    gpc_iterative_test(int, int);
BOOL   congruent(const POSET*, const POSET*);
#ifdef _WIDTH_2_
void   cpcp_2(POSET*);
#define cpcp(g) cpcp_2(&(g))
#else
CPC    cpcp_i(SET, const SET*, const SET*);
#define cpcp(g) (g).t = cpcp_i(FULL_SET, (g).r[0], (g).r[1])
#endif
void   poset_print(const POSET*);

#endif /* _MAIN_H_ */
