/*
 * Minimum Comparison Sorting --- Experiments
 *
 * sort.c
 * Posets, hash tables, Wells's method, sortable functions
 */

/* Copyright 2002-2005, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "main.h"
#if THREADS_NUMBER > 1
#include <pthread.h>

static pthread_mutex_t input_mutex     = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t hash_mutex      = PTHREAD_MUTEX_INITIALIZER;
#ifdef _RW_ALG_
static pthread_cond_t  hash_cond       = PTHREAD_COND_INITIALIZER;
static int             readers_counter = 0;
#endif
static FILE_HEADER     out_fh;
#endif

#ifdef _WIDTH_2_
static BOOL sortable(const POSET*, CPC, int);
static BOOL sortable_2(const POSET*, CPC, int);
#else
static BOOL sortable(const POSET*, CPC);
static BOOL sortable_2(const POSET*, CPC);
#endif

       int   file[3] = {-1, -1, -1};
static int   current_file, current_poset, current_buffered;
       POSET read_buffer[READ_INTERVAL];

/* Tables speed up counting hash functions */
static INDEX h1_tbl[N][N];
static INDEX s1_tbl[N][N];

#ifdef _USE_POF2_TABLE_
/* One element set - Power OF 2 table */
SET pof2_tbl[N];
#endif

#ifdef _USE_MIN_ELEMENT_TABLE_
/* Lowest element in set */
MIN_EL_TBL_EL min_element_tbl[MET_SIZE + 384];
#endif

#ifdef _USE_CARDINALITY_TABLE_
/* Set cardinality - counting ones */
CARD_TBL_EL cardinality_tbl[CAT_SIZE + 384];
#endif

/* Hash tables and stacks for posets storage */
#if THREADS_NUMBER > 1
#define MAX_POSETS_STACK   MAX_STACK
#else
#define MAX_POSETS_STACK   (MAX_STACK + N*N)
#endif
#define MAX_SORTABLE       (MAX_STACK / 2)
#define EMPTY_POS          -1
static INDEX  new_posets_count;
static INDEX  max_posets_count;
static INDEX  sav_posets_count;
static INDEX  hash_tbl[HASH_TBL_SIZE];
static POSET  posets_stack[MAX_POSETS_STACK];
static POSET* const sortable_stack = posets_stack +
                                     (MAX_POSETS_STACK - MAX_SORTABLE);
static INDEX  sortable_count;
static INDEX  sort_tbl[SORT_TBL_SIZE];

#ifdef _WIDTH_2_
/* Fibonacci numbers */
CPC F[N + 3];

static int ilog2(CPC t) {
  CPC t0;
  int i;

  t0 = 1;
  for (i = 0; i < N; i++) {
    if (t0 == t)
      return i;
    t0 *= 2;
  }
  fatal_error("ilog2 not found " CPC_PRINT_FORMAT, t);
  return 0;
}
#endif

void sort_module_init(void) {
  INDEX w1[(N + 1) * (N + 1) / 4];
  int   i, m1, m2;

  #ifdef _USE_POF2_TABLE_
  SET p;

  for (i = 0, p = 1; i < N; i++, p <<= 1)
    pof2_tbl[i] = p;
  #endif

  #ifdef _USE_MIN_ELEMENT_TABLE_
  min_element_tbl[0] = 0;
  min_element_tbl[1] = 0;
  i = 1;
  while (i < MET_SIZE) {
    i++;
    /* i is even */
    min_element_tbl[i] = min_element_tbl[i >> 1] + 1;
    i++;
    /* i is odd */
    min_element_tbl[i] = 0;
  }
  #endif

  #ifdef _USE_CARDINALITY_TABLE_
  cardinality_tbl[0] = 0;
  cardinality_tbl[1] = 1;
  i = 1;
  while (i < CAT_SIZE) {
    i++;
    /* i is even */
    cardinality_tbl[i] = cardinality_tbl[i >> 1];
    i++;
    /* i is odd */
    cardinality_tbl[i] = cardinality_tbl[i >> 1] + 1;
  }
  #endif

  w1[0] = 1;
  for (i = 1; i < (N + 1) * (N + 1) / 4; i++)
    w1[i] = (w1[i - 1] * N) % HASH_TBL_SIZE;
  for (m2 = 0; m2 < N; m2++)
    for (m1 = 0; m1 <= m2; m1++)
      h1_tbl[m1][m2] = h1_tbl[m2][m1] = w1[m1 * (N - m1) + m2];

  w1[0] = 1;
  for (i = 1; i < (N + 1) * (N + 1) / 4; i++)
    w1[i] = (w1[i - 1] * N) % SORT_TBL_SIZE;
  for (m2 = 0; m2 < N; m2++)
    for (m1 = 0; m1 <= m2; m1++)
      s1_tbl[m1][m2] = s1_tbl[m2][m1] = w1[m1 * (N - m1) + m2];

  #ifdef _WIDTH_2_
  F[0] = 0;
  F[1] = 1;
  for (i = 2; i < N + 3; i++)
    F[i] = F[i - 1] + F[i - 2];
  #endif
}

void sort_module_clean(void) {
}

#define EMPTY_POSETS_STACK    1
#define EMPTY_SORTABLE_STACK  2

static void empty_storage(int what) {
  INDEX i;

  if (what & EMPTY_POSETS_STACK) {
    sav_posets_count = max_posets_count = new_posets_count = 0;
    for (i = 0; i < HASH_TBL_SIZE; i++)
      hash_tbl[i] = EMPTY_POS;
  }
  if (what & EMPTY_SORTABLE_STACK) {
    sortable_count = 0;
    for (i = 0; i < SORT_TBL_SIZE; i++)
      sort_tbl[i] = EMPTY_POS;
  }
}

void reset_read_buffer(int file_idx) {
  current_file = file_idx;
  current_poset = 0;
  current_buffered = 0;
}

POSET* get_next_poset(POSET *p) {
  long l;

  #if THREADS_NUMBER > 1
  pthread_mutex_lock(&input_mutex);
  #endif
  current_poset++;
  if (current_poset >= current_buffered) {
    l = read(file[current_file], read_buffer, READ_INTERVAL * sizeof(POSET));
    if (l <= 0) {
      if (p == NULL)
        fatal_error("get_next_poset, read from file %d failed", current_file);
      #if THREADS_NUMBER > 1
      pthread_mutex_unlock(&input_mutex);
      return NULL;
      #endif
    }
    else if (l % sizeof(POSET)) {
      fatal_error("get_next_poset, %ld bytes read from file %d",
                  l, current_file);
    }
    else {
      current_buffered = l / sizeof(POSET);
      current_poset = 0;
    }
  }
  if (p)
    *p = read_buffer[current_poset];
  else
    p = read_buffer + current_poset;
  #if THREADS_NUMBER > 1
  pthread_mutex_unlock(&input_mutex);
  #endif
  return p;
}

static INDEX search(POSET *g) {
  INDEX h;
  int   i;

  h = 0;
  for (i = 0; i < N; i++)
    h += h1_tbl[cardinality(g->r[1][i])][cardinality(g->r[0][i])];
  h %= HASH_TBL_SIZE;
  g->hash = h;
  #if THREADS_NUMBER > 1
  pthread_mutex_lock(&hash_mutex);
  #ifdef _RW_ALG_
  readers_counter++;
  pthread_mutex_unlock(&hash_mutex);
  #endif
  #endif
  for (h = hash_tbl[h]; h != EMPTY_POS; h = posets_stack[h].hash) {
    if (posets_stack[h].n == g->n) {
      if (congruent(posets_stack + h, g)) {
        #if THREADS_NUMBER > 1
        #ifdef _RW_ALG_
        pthread_mutex_lock(&hash_mutex);
        readers_counter--;
        if (readers_counter == 0)
          pthread_cond_signal(&hash_cond);
        #endif
        pthread_mutex_unlock(&hash_mutex);
        #endif
        return h;
      }
    }
  }
  #if THREADS_NUMBER > 1
  #ifdef _RW_ALG_
  pthread_mutex_lock(&hash_mutex);
  readers_counter--;
  if (readers_counter == 0)
    pthread_cond_signal(&hash_cond);
  #endif
  pthread_mutex_unlock(&hash_mutex);
  #endif
  return EMPTY_POS;
}

static void insert(const POSET *g) {
  #if THREADS_NUMBER > 1
  long  l;
  #endif
  INDEX h;
  int   i;

  #if THREADS_NUMBER > 1
  pthread_mutex_lock(&hash_mutex);
  #ifdef _RW_ALG_
  while (readers_counter)
    pthread_cond_wait(&hash_cond, &hash_mutex);
  #endif
  /*
  h = g->hash;
  for (h = hash_tbl[h]; h != EMPTY_POS; h = posets_stack[h].hash) {
    if (posets_stack[h].n == g->n) {
      if (congruent(posets_stack + h, g)) {
        if (pthread_mutex_unlock(&output_mutex))
          fatal_error("insert, output mutex unlock(1) failed");
        return;
      }
    }
  }
  */
  #endif

  if (new_posets_count < max_posets_count) {
    /* Delete an old poset to make space for a new one. */
    h = 0;
    for (i = 0; i < N; i++)
      h += h1_tbl[cardinality(posets_stack[new_posets_count].r[1][i])]
                 [cardinality(posets_stack[new_posets_count].r[0][i])];
    h %= HASH_TBL_SIZE;
    if (hash_tbl[h] == new_posets_count) {
      hash_tbl[h] = posets_stack[new_posets_count].hash;
    }
    else {
      for (h = hash_tbl[h];
           posets_stack[h].hash != new_posets_count;
           h = posets_stack[h].hash);
      posets_stack[h].hash = posets_stack[new_posets_count].hash;
    }
  }
  else {
    /* There is a free position for this poset. */
    max_posets_count = new_posets_count + 1;
  }

  posets_stack[new_posets_count] = *g;
  posets_stack[new_posets_count].hash = hash_tbl[g->hash];
  hash_tbl[g->hash] = new_posets_count;
  new_posets_count++;

  #if THREADS_NUMBER > 1
  if (new_posets_count % UPDATE_INTERVAL == 0 ||
      new_posets_count >= MAX_STACK) {
    l = (new_posets_count - sav_posets_count) * sizeof(POSET);
    if (write(file[OUTF], posets_stack + sav_posets_count, l) != l)
      fatal_error("insert, write failed");
    out_fh.posets_count += new_posets_count - sav_posets_count;
    if (new_posets_count >= MAX_STACK)
      new_posets_count = 0;
    sav_posets_count = new_posets_count;
  }
  pthread_mutex_unlock(&hash_mutex);
  #endif
}

static INDEX sortable_search(POSET *g) {
  INDEX h;
  int   i;

  h = 0;
  for (i = 0; i < N; i++)
    h += s1_tbl[cardinality(g->r[1][i])][cardinality(g->r[0][i])];
  h %= SORT_TBL_SIZE;
  g->hash = h;
  for (h = sort_tbl[h]; h != EMPTY_POS; h = sortable_stack[h].hash)
    if (sortable_stack[h].n == g->n)
      if (congruent(sortable_stack + h, g))
        return h;
  return EMPTY_POS;
}

static void sortable_insert(POSET *g) {
  if (sortable_search(g) == EMPTY_POS) {
    if (sortable_count >= MAX_SORTABLE)
      fatal_error("sortable_insert, hash table overflow");
    sortable_stack[sortable_count] = *g;
    sortable_stack[sortable_count].hash = sort_tbl[g->hash];
    sort_tbl[g->hash] = sortable_count;
    sortable_count++;
  }
}

/*
static void poset_add_pair(POSET *g, int v0, int v1) {
  int i, j;
  SET V[2];

  V[0] = pof2(v0) | g->r[1][v0];
  V[1] = pof2(v1) | g->r[0][v1];

  for (j = 0; j < 2; j++)
    for (i = 0; i < N; i++)
      if (in(V[j], i))
        g->r[j][i] |= V[1 - j];
}
*/
static void poset_add_pair(POSET *g, int v1, int v2) {
  int v;
  SET v_set, n_set;

  /* Add pair (v1,v2) to r[0]. All elements till now greater than v2 are now
     greater than v1. If v was lower than v1, then now v is lower than all
     new elements greater than v1. */
  n_set = g->r[0][v1] |= pof2(v2) | g->r[0][v2];
  v_set = g->r[1][v1];
  while (v_set) {
    v = min_element(v_set);
    v_set ^= pof2(v);
    g->r[0][v] |= n_set;
  }

  /* Add pair (v2,v1) to r[1]. All elements till now lower than v1 are now
     lower than v2. If v was greater than v2, then now v is greater than all
     new elements lower than v2. */
  n_set = g->r[1][v2] |= pof2(v1) | g->r[1][v1];
  v_set = g->r[0][v2];
  while (v_set) {
    v = min_element(v_set);
    v_set ^= pof2(v);
    g->r[1][v] |= n_set;
  }
}

/*
 * Macros implementing Knuth's Lemma
 */

#define G1UNSORTABLE(x) v_set = ((x).r[1][j] | pof2(j)) & ~(pof2(j) - 1); \
                        b_set = ~((x).r[0][k] | pof2(k));                 \
                        while (v_set) {                                   \
                          v = min_element(v_set);                         \
                          v_set ^= pof2(v);                               \
                          A[v] &= b_set;                                  \
                        }                                                 \
                        v_set = ((x).r[0][k] | pof2(k)) & ~(pof2(j) - 1); \
                        b_set = ~((x).r[1][j] | pof2(j));                 \
                        while (v_set) {                                   \
                          v = min_element(v_set);                         \
                          v_set ^= pof2(v);                               \
                          A[v] &= b_set;                                  \
                        }                                                 \

#define G2UNSORTABLE(x) v_set = ((x).r[0][j] | pof2(j)) & ~(pof2(j) - 1); \
                        b_set = ~((x).r[1][k] | pof2(k));                 \
                        while (v_set) {                                   \
                          v = min_element(v_set);                         \
                          v_set ^= pof2(v);                               \
                          A[v] &= b_set;                                  \
                        }                                                 \
                        v_set = ((x).r[1][k] | pof2(k)) & ~(pof2(j) - 1); \
                        b_set = ~((x).r[0][j] | pof2(j));                 \
                        while (v_set) {                                   \
                          v = min_element(v_set);                         \
                          v_set ^= pof2(v);                               \
                          A[v] &= b_set;                                  \
                        }                                                 \

/*
 * Main experiment routines
 */

#if THREADS_NUMBER > 1

static void* one_step_thread(void *analysed) {
  POSET g0, g1, g2;
  int   j, k, v;
  SET   A[N - 1], v_set, b_set;

  *(long*)analysed = 0;
  while (get_next_poset(&g0)) {

    /* Try to add 2 compared elements */
    if (g0.n < N - 1) {
      g1 = g0;
      g1.n += 2;
      poset_add_pair(&g1, g0.n + 1, g0.n);
      if (search(&g1) == EMPTY_POS) {
        g1.t = g0.t / 2;
        insert(&g1);
      }
    }

    /* Try to add 1 element */
    if (g0.n < N) {
      v_set = pof2(g0.n) - 1;
      while (v_set) {
        j = min_element(v_set);
        v_set ^= pof2(j);

        g1 = g0;
        g1.n++;
        poset_add_pair(&g1, j, g0.n);
        if (search(&g1) == EMPTY_POS) {
          g2 = g0;
          g2.n++;
          poset_add_pair(&g2, g0.n, j);
          if (search(&g2) == EMPTY_POS) {
            if (cardinality(g0.r[1][j]) >= cardinality(g0.r[0][j])) {
              cpcp(g1);
              g2.t = g0.t - g1.t;
            }
            else {
              cpcp(g2);
              g1.t = g0.t - g2.t;
            }
            if (g1.t <= out_fh.cut_off && g2.t <= out_fh.cut_off) {
              if (g1.t >= g2.t) {
                insert(&g1);
              }
              else {
                insert(&g2);
              }
            }
            else {
              if (g1.t >= g2.t)
                v_set &= ~g0.r[1][j];
              else
                v_set &= ~g0.r[0][j];
            }
          }
        }
      }
    }

    /* Try to add comparision without new element */
    for (j = 0; j < g0.n - 1; j++)
      A[j] = ~(g0.r[0][j] | g0.r[1][j]) & (FULL_SET >> (N - g0.n)) &
             (FULL_SET << (j + 1));
    for (j = 0; j < g0.n - 1; j++) {
      while (A[j]) {
        k = min_element(A[j]);
        A[j] ^= pof2(k);

        g1 = g0;
        poset_add_pair(&g1, j, k);
        if (search(&g1) == EMPTY_POS) {
          g2 = g0;
          poset_add_pair(&g2, k, j);
          if (search(&g2) == EMPTY_POS) {
            if (cardinality(g0.r[1][j]) + cardinality(g0.r[0][k]) >=
                cardinality(g0.r[0][j]) + cardinality(g0.r[1][k])) {
              cpcp(g1);
              g2.t = g0.t - g1.t;
            }
            else {
              cpcp(g2);
              g1.t = g0.t - g2.t;
            }
            if (g1.t <= out_fh.cut_off && g2.t <= out_fh.cut_off) {
              if (g1.t >= g2.t) {
                insert(&g1);
              }
              else {
                insert(&g2);
              }
            }
            else if (g1.t >= g2.t) {
              G1UNSORTABLE(g0)
            }
            else {
              G2UNSORTABLE(g0)
            }
          }
        }
      }
    }
    (*(long*)analysed)++;
  }
  return NULL;
}

int one_step(BOOL recovery) {
  pthread_t   tid[THREADS_NUMBER  - 1];
  FILE_HEADER in_fh;
  time_t      timer;
  long        analysed[THREADS_NUMBER], l;
  int         i;

  if (recovery)
    fatal_error("one_step, multithreading recovery not implemented");

  timer = time(NULL);
  empty_storage(EMPTY_POSETS_STACK);

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("one_step, read file header failed");
  if (in_fh.version != FILE_VERSION)
    fatal_error("one_step, missing input file version %d", in_fh.version);

  out_fh.version        = FILE_VERSION;
  out_fh.step           = in_fh.step + 1;
  out_fh.analysed_count = 0;
  out_fh.posets_count   = 0;
  out_fh.time           = 0;
  out_fh.cut_off        = in_fh.cut_off / 2;
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("one_step, write(1) file header failed");

  reset_read_buffer(INF);

  for (i = 0; i < THREADS_NUMBER - 1; i++)
    if (pthread_create(&tid[i], NULL, &one_step_thread, &analysed[i]))
      fatal_error("one_step, create thread %d failed", i);
  one_step_thread(&analysed[THREADS_NUMBER - 1]);
  for (i = 0; i < THREADS_NUMBER - 1; i++)
    if (pthread_join(tid[i], NULL))
      fatal_error("one_step, join thread %d failed", i);

  l = (new_posets_count - sav_posets_count) * sizeof(POSET);
  if (l > 0) {
    if (write(file[OUTF], posets_stack + sav_posets_count, l) != l)
      fatal_error("one_step, write failed");
    out_fh.posets_count += new_posets_count - sav_posets_count;
  }
  for (i = 0; i < THREADS_NUMBER; i++)
    out_fh.analysed_count += analysed[i];
  out_fh.time = time(NULL) - timer;

  if (lseek(file[OUTF], 0, SEEK_SET) < 0)
    fatal_error("one_step, lseek(0) failed");
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("one_step, write(2) file header failed");
  l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
  if (lseek(file[OUTF], l, SEEK_SET) != l)
    fatal_error("one_step, lseek(end) failed");

  printf("step %d  input posets %ld  output posets %ld  "
         "computation time %ld\n",
         out_fh.step, out_fh.analysed_count,
         out_fh.posets_count, out_fh.time);
  if (out_fh.analysed_count != in_fh.posets_count)
    fatal_error("one_step, non-analysed posets %ld",
                in_fh.posets_count - out_fh.analysed_count);
  return out_fh.posets_count == 0;
}

static void* alg_search_thread(void *analysed) {
  POSET g;
  #ifdef _WIDTH_2_
  int   c;

  c = ilog2(out_fh.cut_off);
  #endif
  *(long*)analysed = 0;
  while (get_next_poset(&g)) {
    if (search(&g) == EMPTY_POS)
      #ifdef _WIDTH_2_
      if (sortable(&g, out_fh.cut_off, c))
      #else
      if (sortable(&g, out_fh.cut_off))
      #endif
        insert(&g);
    (*(long*)analysed)++;
  }
  return NULL;
}

int alg_search(BOOL recovery) {
  pthread_t   tid[THREADS_NUMBER  - 1];
  FILE_HEADER in_fh;
  time_t      timer;
  long        analysed[THREADS_NUMBER], l;
  int         i;

  if (recovery)
    fatal_error("alg_search, multithreading recovery not implemented");

  timer = time(NULL);
  empty_storage(EMPTY_POSETS_STACK);

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search, read file header failed");
  if (in_fh.version != FILE_VERSION)
    fatal_error("alg_search, missing input file version %d", in_fh.version);

  out_fh.version        = FILE_VERSION;
  out_fh.step           = in_fh.step;
  out_fh.analysed_count = 0;
  out_fh.posets_count   = 0;
  out_fh.time           = 0;
  out_fh.cut_off        = in_fh.cut_off;
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search, write(1) file header failed");

  reset_read_buffer(INF);

  for (i = 0; i < THREADS_NUMBER - 1; i++)
    if (pthread_create(&tid[i], NULL, &alg_search_thread, &analysed[i]))
      fatal_error("alg_search, create thread %d failed", i);
  alg_search_thread(&analysed[THREADS_NUMBER - 1]);
  for (i = 0; i < THREADS_NUMBER - 1; i++)
    if (pthread_join(tid[i], NULL))
      fatal_error("alg_search, join thread %d failed", i);

  l = (new_posets_count - sav_posets_count) * sizeof(POSET);
  if (l > 0) {
    if (write(file[OUTF], posets_stack + sav_posets_count, l) != l)
      fatal_error("alg_search, write failed");
    out_fh.posets_count += new_posets_count - sav_posets_count;
  }
  for (i = 0; i < THREADS_NUMBER; i++)
    out_fh.analysed_count += analysed[i];
  out_fh.time = time(NULL) - timer;

  if (lseek(file[OUTF], 0, SEEK_SET) < 0)
    fatal_error("alg_search, lseek(0) failed");
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search, write(2) file header failed");
  l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
  if (lseek(file[OUTF], l, SEEK_SET) != l)
    fatal_error("alg_search, lseek(end) failed");

  printf("step %d  input posets %ld  sortable posets %ld  "
         "computation time %ld\n",
         out_fh.step, out_fh.analysed_count,
         out_fh.posets_count, out_fh.time);
  if (out_fh.analysed_count != in_fh.posets_count)
    fatal_error("alg_search, non-analysed posets %ld",
                in_fh.posets_count - out_fh.analysed_count);
  return out_fh.posets_count == 0;
}

static void* alg_search_2_thread(void *analysed) {
  POSET g;
  #ifdef _WIDTH_2_
  int   c;

  c = ilog2(out_fh.cut_off);
  #endif
  *(long*)analysed = 0;
  while (get_next_poset(&g)) {
    if (search(&g) == EMPTY_POS)
      #ifdef _WIDTH_2_
      if (sortable_2(&g, out_fh.cut_off, c))
      #else
      if (sortable_2(&g, out_fh.cut_off))
      #endif
        insert(&g);
    (*(long*)analysed)++;
  }
  return NULL;
}

int alg_search_2(BOOL recovery) {
  pthread_t   tid[THREADS_NUMBER  - 1];
  FILE_HEADER in_fh, sort_fh;
  POSET       g;
  time_t      timer;
  long        analysed[THREADS_NUMBER], l;
  int         i;

  if (recovery)
    fatal_error("alg_search_2, multithreading recovery not implemented");

  timer = time(NULL);
  empty_storage(EMPTY_POSETS_STACK | EMPTY_SORTABLE_STACK);

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_2, read input file header failed");
  if (in_fh.version != FILE_VERSION)
    fatal_error("alg_search_2, missing input file version %d",
                in_fh.version);

  if (read(file[SORTF], &sort_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_2, read sortable file header failed");
  if (sort_fh.version != FILE_VERSION)
    fatal_error("alg_search_2, missing sortable file version %d",
                sort_fh.version);

  out_fh.version        = FILE_VERSION;
  out_fh.step           = in_fh.step;
  out_fh.analysed_count = 0;
  out_fh.posets_count   = 0;
  out_fh.time           = 0;
  out_fh.cut_off        = in_fh.cut_off;
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_2, write output file header failed");

  reset_read_buffer(SORTF);
  for (l = 0; l < sort_fh.posets_count; l++) {
    if (get_next_poset(&g))
      sortable_insert(&g);
    else
      fatal_error("alg_search_2, read poset %ld failed", l);
  }

  reset_read_buffer(INF);

  for (i = 0; i < THREADS_NUMBER - 1; i++)
    if (pthread_create(&tid[i], NULL, &alg_search_2_thread, &analysed[i]))
      fatal_error("alg_search_2, create thread %d failed", i);
  alg_search_2_thread(&analysed[THREADS_NUMBER - 1]);
  for (i = 0; i < THREADS_NUMBER - 1; i++)
    if (pthread_join(tid[i], NULL))
      fatal_error("alg_search_2, join thread %d failed", i);

  l = (new_posets_count - sav_posets_count) * sizeof(POSET);
  if (l > 0) {
    if (write(file[OUTF], posets_stack + sav_posets_count, l) != l)
      fatal_error("alg_search_2, write failed");
    out_fh.posets_count += new_posets_count - sav_posets_count;
  }
  for (i = 0; i < THREADS_NUMBER; i++)
    out_fh.analysed_count += analysed[i];
  out_fh.time = time(NULL) - timer;

  if (lseek(file[OUTF], 0, SEEK_SET) < 0)
    fatal_error("alg_search_2, lseek(0) failed");
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_2, write(2) file header failed");
  l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
  if (lseek(file[OUTF], l, SEEK_SET) != l)
    fatal_error("alg_search_2, lseek(end) failed");

  printf("step %d  input posets %ld  sortable posets %ld  "
         "computation time %ld\n",
         out_fh.step, out_fh.analysed_count,
         out_fh.posets_count, out_fh.time);
  if (out_fh.analysed_count != in_fh.posets_count)
    fatal_error("alg_search_2, non-analysed posets %ld",
                in_fh.posets_count - out_fh.analysed_count);
  return out_fh.posets_count == 0;
}

#else /* THREADS_NUMBER > 1 */

int one_step(BOOL recovery) {
  FILE_HEADER in_fh, out_fh;
  POSET       *p_g, g1, g2;
  long        l;
  int         j, k, v;
  SET         A[N - 1], v_set, b_set;

  empty_storage(EMPTY_POSETS_STACK);

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("one_step, read file header failed");
  if (in_fh.version != FILE_VERSION)
    fatal_error("one_step, missing input file version %d", in_fh.version);

  if (recovery) {
    if (read(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("one_step recovery, read output file header failed");
    if (out_fh.version != FILE_VERSION ||
        in_fh.cut_off != out_fh.cut_off * 2 ||
        in_fh.step + 1 != out_fh.step)
      fatal_error("one_step recovery, missing output file");
    reset_read_buffer(OUTF);
    for (l = 0; l < out_fh.posets_count; l++) {
      #ifdef _PRINT_PROGRESS_
      printf("%9ld %9ld\r", out_fh.analysed_count, l);
      #endif
      p_g = get_next_poset(NULL);
      if (search(p_g) == EMPTY_POS) {
        insert(p_g);
        if (new_posets_count >= MAX_STACK)
          new_posets_count = 0;
      }
    }
    sav_posets_count = new_posets_count;
    l = sizeof(FILE_HEADER) + out_fh.analysed_count * sizeof(POSET);
    if (lseek(file[INF], l, SEEK_SET) != l)
      fatal_error("one_step recovery, lseek(input file) failed");
    l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
    if (lseek(file[OUTF], l, SEEK_SET) != l)
      fatal_error("one_step recovery, lseek(output file) failed");
  }
  else {
    memcpy(&out_fh, &in_fh, sizeof(FILE_HEADER));
    /* out_fh.cut_off = 2^{c_max - out_fh.step} */
    out_fh.cut_off /= 2;
    out_fh.step++;
    out_fh.analysed_count = 0;
    out_fh.posets_count = 0;
    out_fh.time = 0;
    if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("one_step, write output file header failed");
  }
  #ifdef _PRINT_PROGRESS_
  printf("%9ld %9ld\r", out_fh.analysed_count, out_fh.posets_count);
  #endif

  /* in_fh.time is reused as analysis start time stamp and
     out_fh.time is elapsed analysis time */
  in_fh.time = time(NULL) - out_fh.time;

  reset_read_buffer(INF);
  while (out_fh.analysed_count < in_fh.posets_count) {
    p_g = get_next_poset(NULL);

    /* Try to add 2 compared elements */
    if (p_g->n < N - 1) {
      g1 = *p_g;
      g1.n += 2;
      poset_add_pair(&g1, p_g->n + 1, p_g->n);
      if (search(&g1) == EMPTY_POS) {
        g1.t = p_g->t / 2;
        insert(&g1);
      }
    }

    /* Try to add 1 element */
    if (p_g->n < N) {
      v_set = pof2(p_g->n) - 1;
      while (v_set) {
        j = min_element(v_set);
        v_set ^= pof2(j);

        g1 = *p_g;
        g1.n++;
        poset_add_pair(&g1, j, p_g->n);
        if (search(&g1) == EMPTY_POS) {
          g2 = *p_g;
          g2.n++;
          poset_add_pair(&g2, p_g->n, j);
          if (search(&g2) == EMPTY_POS) {
            if (cardinality(p_g->r[1][j]) >= cardinality(p_g->r[0][j])) {
              cpcp(g1);
              g2.t = p_g->t - g1.t;
            }
            else {
              cpcp(g2);
              g1.t = p_g->t - g2.t;
            }
            if (g1.t <= out_fh.cut_off && g2.t <= out_fh.cut_off) {
              if (g1.t >= g2.t) {
                insert(&g1);
              }
              else {
                insert(&g2);
              }
            }
            else {
              if (g1.t >= g2.t)
                v_set &= ~p_g->r[1][j];
              else
                v_set &= ~p_g->r[0][j];
            }
          }
        }
      }
    }

    /* Try to add comparision without new element */
    for (j = 0; j < p_g->n - 1; j++)
      A[j] = ~(p_g->r[0][j] | p_g->r[1][j]) & (FULL_SET >> (N - p_g->n)) &
             (FULL_SET << (j + 1));
    for (j = 0; j < p_g->n - 1; j++) {
      while (A[j]) {
        k = min_element(A[j]);
        A[j] ^= pof2(k);

        g1 = *p_g;
        poset_add_pair(&g1, j, k);
        if (search(&g1) == EMPTY_POS) {
          g2 = *p_g;
          poset_add_pair(&g2, k, j);
          if (search(&g2) == EMPTY_POS) {
            if (cardinality(p_g->r[1][j]) + cardinality(p_g->r[0][k]) >=
                cardinality(p_g->r[0][j]) + cardinality(p_g->r[1][k])) {
              cpcp(g1);
              g2.t = p_g->t - g1.t;
            }
            else {
              cpcp(g2);
              g1.t = p_g->t - g2.t;
            }
            if (g1.t <= out_fh.cut_off && g2.t <= out_fh.cut_off) {
              if (g1.t >= g2.t) {
                insert(&g1);
              }
              else {
                insert(&g2);
              }
            }
            else if (g1.t >= g2.t) {
              G1UNSORTABLE(*p_g)
            }
            else {
              G2UNSORTABLE(*p_g)
            }
          }
        }
      }
    }
    out_fh.analysed_count++;
    if (out_fh.analysed_count % UPDATE_INTERVAL == 0 ||
        out_fh.analysed_count == in_fh.posets_count ||
        new_posets_count >= MAX_STACK) {
      l = (new_posets_count - sav_posets_count) * sizeof(POSET);
      if (write(file[OUTF], posets_stack + sav_posets_count, l) != l)
        fatal_error("one_step save, write posets failed");
      if (lseek(file[OUTF], 0, SEEK_SET) < 0)
        fatal_error("one_step save, lseek(0) failed");
      out_fh.posets_count += new_posets_count - sav_posets_count;
      out_fh.time = time(NULL) - in_fh.time;
      if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) !=
          sizeof(FILE_HEADER))
        fatal_error("one_step save, write file header failed");
      l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
      if (lseek(file[OUTF], l, SEEK_SET) != l)
        fatal_error("one_step save, lseek(end) failed");
      if (new_posets_count >= MAX_STACK)
        new_posets_count = 0;
      sav_posets_count = new_posets_count;
    }
    #ifdef _PRINT_PROGRESS_
    printf("%9ld %9ld\r", out_fh.analysed_count,
           out_fh.posets_count + new_posets_count - sav_posets_count);
    #endif
  }

  printf("step %d  input posets %ld  output posets %ld  "
         "computation time %ld\n",
         out_fh.step, out_fh.analysed_count,
         out_fh.posets_count, out_fh.time);
  return out_fh.posets_count == 0;
}

int alg_search(BOOL recovery) {
  FILE_HEADER in_fh, out_fh;
  POSET       *p_g;
  long        l;
  #ifdef _WIDTH_2_
  int         c;
  #endif

  empty_storage(EMPTY_POSETS_STACK);

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search, read input file header failed");
  if (in_fh.version != FILE_VERSION)
    fatal_error("alg_search, missing input file version %d", in_fh.version);

  if (recovery) {
    if (read(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("alg_search recovery, read output file header failed");
    if (out_fh.version != FILE_VERSION ||
        in_fh.cut_off != out_fh.cut_off ||
        in_fh.step != out_fh.step)
      fatal_error("alg_search recovery, missing output file");
    reset_read_buffer(OUTF);
    for (l = 0; l < out_fh.posets_count; l++) {
      #ifdef _PRINT_PROGRESS_
      printf("%9ld %9ld\r", out_fh.analysed_count, l);
      #endif
      p_g = get_next_poset(NULL);
      if (search(p_g) != EMPTY_POS)
        fatal_error("alg_search recovery, congruent poset found");
      else
        insert(p_g);
    }
    l = sizeof(FILE_HEADER) + out_fh.analysed_count * sizeof(POSET);
    if (lseek(file[INF], l, SEEK_SET) != l)
      fatal_error("alg_search recovery, lseek(input file) failed");
    l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
    if (lseek(file[OUTF], l, SEEK_SET) != l)
      fatal_error("alg_search recovery, lseek(output file) failed");
  }
  else {
    memcpy(&out_fh, &in_fh, sizeof(FILE_HEADER));
    out_fh.analysed_count = 0;
    out_fh.posets_count = 0;
    out_fh.time = 0;
    if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("alg_search: can't write output file header");
  }
  #ifdef _PRINT_PROGRESS_
  printf("%9ld %9ld\r", out_fh.analysed_count, (long)new_posets_count);
  #endif

  /* in_fh.time is reused as analysis start time stamp and
     out_fh.time is elapsed analysis time */
  in_fh.time = time(NULL) - out_fh.time;

  #ifdef _WIDTH_2_
  c = ilog2(out_fh.cut_off);
  #endif
  reset_read_buffer(INF);
  while (out_fh.analysed_count < in_fh.posets_count &&
         out_fh.posets_count < MAX_STACK) {
    p_g = get_next_poset(NULL);

    /* result = (g0.t <= cut_off) ? sortable(&g0, out_fh.cut_off) : FALSE; */
    /* Here linear extensions count is not greather than cut_off. */
    if (search(p_g) == EMPTY_POS)
      #ifdef _WIDTH_2_
      if (sortable(p_g, out_fh.cut_off, c))
      #else
      if (sortable(p_g, out_fh.cut_off))
      #endif
        insert(p_g);

    out_fh.analysed_count++;
    if (out_fh.analysed_count % UPDATE_INTERVAL == 0 ||
        out_fh.analysed_count == in_fh.posets_count ||
        new_posets_count >= MAX_STACK) {
      l = (new_posets_count - out_fh.posets_count) * sizeof(POSET);
      if (l > 0)
        if (write(file[OUTF], posets_stack + out_fh.posets_count, l) != l)
          fatal_error("alg_search save, write posets failed");
      if (lseek(file[OUTF], 0, SEEK_SET) < 0)
        fatal_error("alg_search save, lseek(0) failed");
      out_fh.posets_count = new_posets_count;
      out_fh.time = time(NULL) - in_fh.time;
      if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) !=
          sizeof(FILE_HEADER))
        fatal_error("alg_search save, write file header failed");
      l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
      if (lseek(file[OUTF], l, SEEK_SET) != l)
        fatal_error("alg_search save, lseek(end) failed");
    }
    #ifdef _PRINT_PROGRESS_
    printf("%9ld %9ld\r", out_fh.analysed_count, (long)new_posets_count);
    #endif
  }
  printf("step %d  input posets %ld  sortable posets %ld  "
         "computation time %ld\n",
         out_fh.step, out_fh.analysed_count,
         out_fh.posets_count, out_fh.time);
  if (out_fh.analysed_count != in_fh.posets_count)
    fatal_error("alg_search, non-analysed posets %ld",
                in_fh.posets_count - out_fh.analysed_count);
  return out_fh.posets_count == 0;
}

int alg_search_2(BOOL recovery) {
  FILE_HEADER in_fh, out_fh, sort_fh;
  POSET       *p_g;
  long        l;
  #ifdef _WIDTH_2_
  int         c;
  #endif

  empty_storage(EMPTY_POSETS_STACK | EMPTY_SORTABLE_STACK);

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_2, read input file header failed");
  if (in_fh.version != FILE_VERSION)
    fatal_error("alg_search_2, missing input file version %d",
                in_fh.version);

  if (read(file[SORTF], &sort_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_2, read sortable file header failed");
  if (sort_fh.version != FILE_VERSION)
    fatal_error("alg_search_2, missing sortable file version %d",
                sort_fh.version);
  reset_read_buffer(SORTF);

  if (recovery) {
    for (l = 0; l < sort_fh.posets_count; l++) {
      #ifdef _PRINT_PROGRESS_
      printf("%9ld %9ld\r", sort_fh.analysed_count, l);
      #endif
      sortable_insert(get_next_poset(NULL));
    }
    if (read(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("alg_search_2 recovery, read output file header failed");
    if (out_fh.version != FILE_VERSION ||
        in_fh.cut_off != out_fh.cut_off ||
        in_fh.step != out_fh.step)
      fatal_error("alg_search_2 recovery, missing output file");
    reset_read_buffer(OUTF);
    for (l = 0; l < out_fh.posets_count; l++) {
      #ifdef _PRINT_PROGRESS_
      printf("%9ld %9ld\r", out_fh.analysed_count, l);
      #endif
      p_g = get_next_poset(NULL);
      if (search(p_g) != EMPTY_POS)
        fatal_error("alg_search_2 recovery, congruent poset found");
      else
        insert(p_g);
    }
    l = sizeof(FILE_HEADER) + out_fh.analysed_count * sizeof(POSET);
    if (lseek(file[INF], l, SEEK_SET) != l)
      fatal_error("alg_search_2 recovery, lseek(input file) failed");
    l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
    if (lseek(file[OUTF], l, SEEK_SET) != l)
      fatal_error("alg_search_2 recovery, lseek(output file) failed");
  }
  else {
    memcpy(&out_fh, &in_fh, sizeof(FILE_HEADER));
    out_fh.analysed_count = 0;
    out_fh.posets_count = 0;
    out_fh.time = 0;
    if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("alg_search_2, write output file header failed");
    for (l = 0; l < sort_fh.posets_count; l++) {
      #ifdef _PRINT_PROGRESS_
      printf("%9ld %9ld\r", sort_fh.analysed_count, l);
      #endif
      sortable_insert(get_next_poset(NULL));
    }
  }
  #ifdef _PRINT_PROGRESS_
  printf("%9ld %9ld\r", out_fh.analysed_count, (long)new_posets_count);
  #endif

  /* in_fh.time is reused as analysis start time stamp and
     out_fh.time is elapsed analysis time */
  in_fh.time = time(NULL) - out_fh.time;

  #ifdef _WIDTH_2_
  c = ilog2(out_fh.cut_off);
  #endif
  reset_read_buffer(INF);
  while (out_fh.analysed_count < in_fh.posets_count &&
         out_fh.posets_count < MAX_STACK - MAX_SORTABLE) {
    p_g = get_next_poset(NULL);

    /* Here linear extensions count is not greather than cut_off. */
    if (search(p_g) == EMPTY_POS)
      #ifdef _WIDTH_2_
      if (sortable_2(p_g, out_fh.cut_off, c))
      #else
      if (sortable_2(p_g, out_fh.cut_off))
      #endif
        insert(p_g);

    out_fh.analysed_count++;
    if (out_fh.analysed_count % UPDATE_INTERVAL == 0 ||
        out_fh.analysed_count == in_fh.posets_count ||
        new_posets_count >= MAX_STACK - MAX_SORTABLE) {
      l = (new_posets_count - out_fh.posets_count) * sizeof(POSET);
      if (l > 0)
        if (write(file[OUTF], posets_stack + out_fh.posets_count, l) != l)
          fatal_error("alg_search_2 save, write posets failed");
      if (lseek(file[OUTF], 0, SEEK_SET) < 0)
        fatal_error("alg_search_2 save, lseek(0) failed");
      out_fh.posets_count = new_posets_count;
      out_fh.time = time(NULL) - in_fh.time;
      if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) !=
          sizeof(FILE_HEADER))
        fatal_error("alg_search_2 save, write file header failed");
      l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
      if (lseek(file[OUTF], l, SEEK_SET) != l)
        fatal_error("alg_search_2 save, lseek(end) failed");
    }
    #ifdef _PRINT_PROGRESS_
    printf("%9ld %9ld\r", out_fh.analysed_count, (long)new_posets_count);
    #endif
  }
  printf("step %d  input posets %ld  sortable posets %ld  "
         "computation time %ld\n",
         out_fh.step, out_fh.analysed_count,
         out_fh.posets_count, out_fh.time);
  if (out_fh.analysed_count != in_fh.posets_count)
    fatal_error("alg_search_2, non-analysed posets %ld",
                in_fh.posets_count - out_fh.analysed_count);
  return out_fh.posets_count == 0;
}

#endif /* THREADS_NUMBER > 1 */

int alg_search_3(int match_type) {
  FILE_HEADER in_fh, out_fh, sort_fh;
  POSET       *p_g, g1, g2;
  long        l;
  INDEX       idx1, idx2;
  int         j, k, n;
  SET         A[N - 1], v_set;

  empty_storage(EMPTY_POSETS_STACK | EMPTY_SORTABLE_STACK);

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_3, read input file header failed");
  if (in_fh.version != FILE_VERSION)
    fatal_error("alg_search_3, missing input file version %d",
                in_fh.version);

  if (read(file[SORTF], &sort_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_3, read sortable file header failed");
  if (sort_fh.version != FILE_VERSION)
    fatal_error("alg_search_3, missing sortable file version %d",
                sort_fh.version);

  if (match_type == 1)
    memcpy(&out_fh, &sort_fh, sizeof(FILE_HEADER));
  else if (match_type == 2)
    memcpy(&out_fh, &in_fh, sizeof(FILE_HEADER));
  else
    fatal_error("alg_search_3, missing match_type %d", match_type);
  out_fh.analysed_count = 0;
  out_fh.posets_count = 0;
  out_fh.time = 0;
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("alg_search_3, write output file header failed");

  reset_read_buffer(SORTF);
  for (l = 0; l < sort_fh.posets_count; l++)
    sortable_insert(get_next_poset(NULL));

  /* in_fh.time is reused as analysis start time stamp and
     out_fh.time is elapsed analysis time */
  in_fh.time = time(NULL) - out_fh.time;

  reset_read_buffer(INF);
  while (out_fh.analysed_count < in_fh.posets_count &&
         out_fh.posets_count < MAX_STACK - MAX_SORTABLE) {
    p_g = get_next_poset(NULL);
    n = p_g->n;

    if (match_type == 2 && p_g->t <= 6) {
      if (search(p_g) == EMPTY_POS)
        insert(p_g);
      goto LOOP_END;
    }

    if (match_type == 2 && n < N - 1) {
      g1 = *p_g;
      g1.n += 2;
      poset_add_pair(&g1, n + 1, n);
      if (sortable_search(&g1) != EMPTY_POS && search(p_g) == EMPTY_POS)
        insert(p_g);
      goto LOOP_END;
    }

    /* Try to add 1 element */
    if (n < N) {
      v_set = pof2(n) - 1;
      while (v_set) {
        j = min_element(v_set);
        v_set ^= pof2(j);

        g1 = *p_g;
        g1.n++;
        poset_add_pair(&g1, j, n);
        g2 = *p_g;
        g2.n++;
        poset_add_pair(&g2, n, j);
        idx1 = sortable_search(&g1);
        idx2 = sortable_search(&g2);
        if (match_type == 1) {
          if (idx1 != EMPTY_POS && idx2 == EMPTY_POS &&
              (g2.t = p_g->t - sortable_stack[idx1].t) <= out_fh.cut_off) {
            if (search(&g2) == EMPTY_POS)
              insert(&g2);
          }
          else if (idx1 == EMPTY_POS && idx2 != EMPTY_POS &&
              (g1.t = p_g->t - sortable_stack[idx2].t) <= out_fh.cut_off) {
            if (search(&g1) == EMPTY_POS)
              insert(&g1);
          }
        }
        else if (match_type == 2) {
          if (idx1 != EMPTY_POS && idx2 != EMPTY_POS) {
            if (search(p_g) == EMPTY_POS)
              insert(p_g);
            goto LOOP_END;
          }
        }
      }
    }

    /* Try to add comparision without new element */
    for (j = 0; j < n - 1; j++)
      A[j] = ~(p_g->r[0][j] | p_g->r[1][j]) & (FULL_SET >> (N - n)) &
             (FULL_SET << (j + 1));
    for (j = 0; j < n - 1; j++) {
      while (A[j]) {
        k = min_element(A[j]);
        A[j] ^= pof2(k);

        g1 = *p_g;
        poset_add_pair(&g1, j, k);
        g2 = *p_g;
        poset_add_pair(&g2, k, j);
        idx1 = sortable_search(&g1);
        idx2 = sortable_search(&g2);
        if (match_type == 1) {
          if (idx1 != EMPTY_POS && idx2 == EMPTY_POS &&
              (g2.t = p_g->t - sortable_stack[idx1].t) <= out_fh.cut_off) {
            if (search(&g2) == EMPTY_POS)
              insert(&g2);
          }
          else if (idx1 == EMPTY_POS && idx2 != EMPTY_POS &&
              (g1.t = p_g->t - sortable_stack[idx2].t) <= out_fh.cut_off) {
            if (search(&g1) == EMPTY_POS)
              insert(&g1);
          }
        }
        else if (match_type == 2) {
          if (idx1 != EMPTY_POS && idx2 != EMPTY_POS) {
            if (search(p_g) == EMPTY_POS)
              insert(p_g);
            goto LOOP_END;
          }
        }
      }
    }

   LOOP_END:
    out_fh.analysed_count++;
    if (out_fh.analysed_count % UPDATE_INTERVAL == 0 ||
        out_fh.analysed_count == in_fh.posets_count ||
        new_posets_count >= MAX_STACK - MAX_SORTABLE) {
      l = (new_posets_count - out_fh.posets_count) * sizeof(POSET);
      if (l > 0)
        if (write(file[OUTF], posets_stack + out_fh.posets_count, l) != l)
          fatal_error("alg_search_3 save, write posets failed");
      if (lseek(file[OUTF], 0, SEEK_SET) < 0)
        fatal_error("alg_search_3 save, lseek(0) failed");
      out_fh.posets_count = new_posets_count;
      out_fh.time = time(NULL) - in_fh.time;
      if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) !=
          sizeof(FILE_HEADER))
        fatal_error("alg_search_3 save, write file header failed");
      l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
      if (lseek(file[OUTF], l, SEEK_SET) != l)
        fatal_error("alg_search_3 save, lseek(end) failed");
    }
  }
  if (match_type == 1)
    printf("step %d  input posets %ld  output posets %ld  "
           "computation time %ld\n",
           out_fh.step, out_fh.analysed_count,
           out_fh.posets_count, out_fh.time);
  else if (match_type == 2)
    printf("step %d  input posets %ld  sortable posets %ld  "
           "computation time %ld\n",
           out_fh.step, out_fh.analysed_count,
           out_fh.posets_count, out_fh.time);
  if (out_fh.analysed_count != in_fh.posets_count)
    fatal_error("alg_search_3, non-analysed posets %ld",
                in_fh.posets_count - out_fh.analysed_count);
  return out_fh.posets_count == 0;
}

#ifdef _WIDTH_2_
static BOOL sortable(const POSET *p_g, CPC cut_off, int c) {
#else
static BOOL sortable(const POSET *p_g, CPC cut_off) {
#endif
  POSET g1, g2;
  int   j, k, n, v;
  SET   A[N - 1], v_set, b_set;

  #ifdef _WIDTH_2_
  if (p_g->t < F[c + 3])
  #else
  if (p_g->t <= 6) /* In the first version here was "t <= 2",       */
  #endif
    return TRUE;   /* "t <= 8" is wrong, "t <= 7" is open question. */

  cut_off /= 2;
  #ifdef _WIDTH_2_
  c--;
  #endif
  n = p_g->n;

  /* Try to add 2 compared elements */
  if (n < N - 1) {
    g1 = *p_g;
    g1.n += 2;
    poset_add_pair(&g1, n + 1, n);
    g1.t = p_g->t / 2;
    #ifdef _WIDTH_2_
    if (sortable(&g1, cut_off, c))
    #else
    if (sortable(&g1, cut_off))
    #endif
      return TRUE;
  }

  /* Try to add 1 element */
  if (n < N) {
    v_set = pof2(n) - 1;
    while (v_set) {
      j = min_element(v_set);
      v_set ^= pof2(j);

      g1 = *p_g;
      g1.n++;
      poset_add_pair(&g1, j, n);
      g2 = *p_g;
      g2.n++;
      poset_add_pair(&g2, n, j);
      if (cardinality(p_g->r[1][j]) >= cardinality(p_g->r[0][j])) {
        cpcp(g1);
        g2.t = p_g->t - g1.t;
      }
      else {
        cpcp(g2);
        g1.t = p_g->t - g2.t;
      }
      if (g1.t <= cut_off && g2.t <= cut_off) {
        if (g1.t >= g2.t) {
          #ifdef _WIDTH_2_
          if (sortable(&g1, cut_off, c)) {
            if (sortable(&g2, cut_off, c))
          #else
          if (sortable(&g1, cut_off)) {
            if (sortable(&g2, cut_off))
          #endif
              return TRUE;
            else
              v_set &= ~p_g->r[0][j];
          }
          else
            v_set &= ~p_g->r[1][j];
        }
        else {
          #ifdef _WIDTH_2_
          if (sortable(&g2, cut_off, c)) {
            if (sortable(&g1, cut_off, c))
          #else
          if (sortable(&g2, cut_off)) {
            if (sortable(&g1, cut_off))
          #endif
              return TRUE;
            else
              v_set &= ~p_g->r[1][j];
          }
          else
            v_set &= ~p_g->r[0][j];
        }
      }
      else if (g1.t >= g2.t)
        v_set &= ~p_g->r[1][j];
      else
        v_set &= ~p_g->r[0][j];
    }
  }

  /* Try to add comparision without new element */
  for (j = 0; j < n - 1; j++)
    A[j] = ~(p_g->r[0][j] | p_g->r[1][j]) & (FULL_SET >> (N - n)) &
           (FULL_SET << (j + 1));
  for (j = 0; j < n - 1; j++) {
    while (A[j]) {
      k = min_element(A[j]);
      A[j] ^= pof2(k);

      g1 = *p_g;
      poset_add_pair(&g1, j, k);
      g2 = *p_g;
      poset_add_pair(&g2, k, j);
      if (cardinality(p_g->r[1][j]) + cardinality(p_g->r[0][k]) >=
          cardinality(p_g->r[0][j]) + cardinality(p_g->r[1][k])) {
        cpcp(g1);
        g2.t = p_g->t - g1.t;
      }
      else {
        cpcp(g2);
        g1.t = p_g->t - g2.t;
      }
      if (g1.t <= cut_off && g2.t <= cut_off) {
        if (g1.t >= g2.t) {
          #ifdef _WIDTH_2_
          if (sortable(&g1, cut_off, c)) {
            if (sortable(&g2, cut_off, c)) {
          #else
          if (sortable(&g1, cut_off)) {
            if (sortable(&g2, cut_off)) {
          #endif
              return TRUE;
            }
            else {
              G2UNSORTABLE(*p_g)
            }
          }
          else {
            G1UNSORTABLE(*p_g)
          }
        }
        else {
          #ifdef _WIDTH_2_
          if (sortable(&g2, cut_off, c)) {
            if (sortable(&g1, cut_off, c)) {
          #else
          if (sortable(&g2, cut_off)) {
            if (sortable(&g1, cut_off)) {
          #endif
              return TRUE;
            }
            else {
              G1UNSORTABLE(*p_g)
            }
          }
          else {
            G2UNSORTABLE(*p_g)
          }
        }
      }
      else if (g1.t >= g2.t) {
        G1UNSORTABLE(*p_g)
      }
      else {
        G2UNSORTABLE(*p_g)
      }
    }
  }

  return FALSE;
}

#ifdef _WIDTH_2_
static BOOL sortable_2(const POSET *p_g, CPC cut_off, int c) {
#else
static BOOL sortable_2(const POSET *p_g, CPC cut_off) {
#endif
  POSET g1, g2;
  INDEX idx1, idx2;
  int   j, k, n, v;
  SET   A[N - 1], v_set, b_set;

  #ifdef _WIDTH_2_
  if (p_g->t < F[c + 3])
  #else
  if (p_g->t <= 6)
  #endif
    return TRUE;

  cut_off /= 2;
  #ifdef _WIDTH_2_
  c--;
  #endif
  n = p_g->n;

  /* Try to add 2 compared elements */
  if (n < N - 1) {
    g1 = *p_g;
    g1.n += 2;
    poset_add_pair(&g1, n + 1, n);
    if (sortable_search(&g1) != EMPTY_POS)
      return TRUE;
  }

  /* Try to add 1 element */
  if (n < N) {
    v_set = pof2(n) - 1;
    while (v_set) {
      j = min_element(v_set);
      v_set ^= pof2(j);

      g1 = *p_g;
      g1.n++;
      poset_add_pair(&g1, j, n);
      g2 = *p_g;
      g2.n++;
      poset_add_pair(&g2, n, j);
      idx1 = sortable_search(&g1);
      idx2 = sortable_search(&g2);
      if (idx1 != EMPTY_POS && idx2 != EMPTY_POS) {
        return TRUE;
      }
      else if (idx1 != EMPTY_POS &&
               (g2.t = p_g->t - sortable_stack[idx1].t) <= cut_off) {
        #ifdef _WIDTH_2_
        if (sortable(&g2, cut_off, c))
        #else
        if (sortable(&g2, cut_off))
        #endif
          return TRUE;
        else
          v_set &= ~p_g->r[0][j];
      }
      else if (idx2 != EMPTY_POS &&
               (g1.t = p_g->t - sortable_stack[idx2].t) <= cut_off) {
        #ifdef _WIDTH_2_
        if (sortable(&g1, cut_off, c))
        #else
        if (sortable(&g1, cut_off))
        #endif
          return TRUE;
        else
          v_set &= ~p_g->r[1][j];
      }
    }
  }

  /* Try to add comparision without new element */
  for (j = 0; j < n - 1; j++)
    A[j] = ~(p_g->r[0][j] | p_g->r[1][j]) & (FULL_SET >> (N - n)) &
           (FULL_SET << (j + 1));
  for (j = 0; j < n - 1; j++) {
    while (A[j]) {
      k = min_element(A[j]);
      A[j] ^= pof2(k);

      g1 = *p_g;
      poset_add_pair(&g1, j, k);
      g2 = *p_g;
      poset_add_pair(&g2, k, j);
      idx1 = sortable_search(&g1);
      idx2 = sortable_search(&g2);
      if (idx1 != EMPTY_POS && idx2 != EMPTY_POS) {
        return TRUE;
      }
      else if (idx1 != EMPTY_POS &&
               (g2.t = p_g->t - sortable_stack[idx1].t) <= cut_off) {
        #ifdef _WIDTH_2_
        if (sortable(&g2, cut_off, c)) {
        #else
        if (sortable(&g2, cut_off)) {
        #endif
          return TRUE;
        }
        else {
          G2UNSORTABLE(*p_g)
        }
      }
      else if (idx2 != EMPTY_POS &&
              (g1.t = p_g->t - sortable_stack[idx2].t) <= cut_off) {
        #ifdef _WIDTH_2_
        if (sortable(&g1, cut_off, c)) {
        #else
        if (sortable(&g1, cut_off)) {
        #endif
          return TRUE;
        }
        else {
          G1UNSORTABLE(*p_g)
        }
      }
    }
  }

  return FALSE;
}

/*
 * Auxiliary routines
 */

int first_step(int c_max, int c) {
  FILE_HEADER fh;
  POSET       g;
  int         i;

  file_header_empty(&fh);
  fh.version        = FILE_VERSION;
  fh.analysed_count = 0;
  fh.posets_count   = c >= 0;
  fh.time           = 0;
  if (c < -c_max)
    c = c_max;
  else if (c < 0)
    c = -c;
  else if (2 * c > N)
    c = N / 2;
  fh.step           = c;
  fh.cut_off        = 1; /* cut_off = 2^{c_max - fh.step} */
  for (i = c; i < c_max; i++)
    fh.cut_off *= 2;
  if (write(file[OUTF], &fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("first_step, write file header failed");

  if (fh.posets_count) {
    poset_empty(&g);
    g.n = 2 * c;
    for (i = 0; i < c; i++)
      poset_add_pair(&g, 2 * i + 1, 2 * i);
    /* g.t = N! / 2^c */
    g.t = 1;
    for (i = 2; i <= N; i++)
      g.t *= i;
    for (i = c; i > 0; i--)
      g.t /= 2;
    if (g.t > fh.cut_off)
      printf("Required poset cannot be sorted using %d comparisons.\n", c_max);
    if (write(file[OUTF], &g, sizeof(POSET)) != sizeof(POSET))
      fatal_error("first_step, write poset failed");
  }

  printf("step %d  input posets %ld  output posets %ld\n",
         fh.step, fh.analysed_count, fh.posets_count);
  return fh.posets_count == 0;
}

int two_chains(int m, int c_max) {
  FILE_HEADER fh;
  POSET       g;
  int         i;

  file_header_empty(&fh);
  fh.version        = FILE_VERSION;
  fh.step           = 0;
  fh.analysed_count = 0;
  fh.posets_count   = 0;
  fh.time           = 0;

  poset_empty(&g);
  g.n = N;
  for (i = 1; i < N; i++)
    if (i != m)
      poset_add_pair(&g, i - 1, i);
  cpcp(g);
  fh.cut_off = 1;
  if (c_max >= 0)
    for (i = 0; i < c_max; i++)
      fh.cut_off *= 2;
  else
    while (g.t > fh.cut_off)
      fh.cut_off *= 2;
  if (g.t > fh.cut_off)
    printf("Chains of lengths %d and %d cannot be merged "
           "using %d comparisons.\n", m, N - m, c_max);
  else
    fh.posets_count = 1;

  if (write(file[OUTF], &fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("two_chains, write file header failed");
  if (fh.posets_count)
    if (write(file[OUTF], &g, sizeof(POSET)) != sizeof(POSET))
      fatal_error("two_chains, write poset failed");

  printf("step %d  input posets %ld  output posets %ld\n",
         fh.step, fh.analysed_count, fh.posets_count);
  return fh.posets_count == 0;
}

int special_poset(int type, int c_max) {
  FILE_HEADER fh;
  POSET       g;
  int         i, j;

  file_header_empty(&fh);
  fh.version        = FILE_VERSION;
  fh.step           = 0;
  fh.analysed_count = 0;
  fh.posets_count   = 0;
  fh.time           = 0;

  poset_empty(&g);
  g.n = N;
  switch (type) {
    case 1: /* hook and chain */
      poset_add_pair(&g, 0, 1);
      poset_add_pair(&g, 2, 3);
      poset_add_pair(&g, 1, 3);
      for (i = 5; i < N; i++)
        poset_add_pair(&g, i - 1, i);
      break;
    case 2: /* L-ladder */
      for (i = 0; i < N; i++) {
        for (j = i + 1; j < N; j++) {
          if (i <= j - 2) {
            poset_add_pair(&g, i, j);
          }
        }
      }
      break;
    case 3: /* M-ladder */
      for (i = 0; i < N; i++) {
        for (j = i + 1; j < N; j++) {
          if ((i <= j - 2 && !(i & 1)) ||
              (i <= j - 3 && (i & 1))) {
            poset_add_pair(&g, i, j);
          }
        }
      }
      break;
    default:
      fatal_error("unknown poset type %d", type);
  }
  cpcp(g);
  fh.cut_off = 1;
  for (i = 0; i < c_max; i++)
    fh.cut_off *= 2;
  if (g.t > fh.cut_off) {
    switch (type) {
      case 1:
        printf("Hook and chain of length %d cannot be merged "
               "using %d comparisons.\n", N - 4, c_max);
        break;
      case 2:
        printf("L-ladder containing %d elements cannot be sorted "
               "using %d comparisons.\n", N, c_max);
        break;
      case 3:
        printf("M-ladder containing %d elements cannot be sorted "
               "using %d comparisons.\n", N, c_max);
        break;
    }
  }
  else
    fh.posets_count = 1;

  if (write(file[OUTF], &fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("special_poset, write file header failed");
  if (fh.posets_count)
    if (write(file[OUTF], &g, sizeof(POSET)) != sizeof(POSET))
      fatal_error("special_poset, write poset failed");

  printf("step %d  input posets %ld  output posets %ld\n",
         fh.step, fh.analysed_count, fh.posets_count);
  return fh.posets_count == 0;
}

int generate_from_description(const char* in_file, int c_max) {
  FILE_HEADER fh;
  POSET       g0, g1;
  FILE*       fd;
  int         c, j, k, count, x[12], y[12];
  CPC         t_max;
  char        buffer0[16], buffer1[16];

  file_header_empty(&fh);
  fh.version        = FILE_VERSION;
  fh.step           = 0;
  fh.analysed_count = 0;
  fh.posets_count   = 0;
  fh.time           = 0;
  poset_empty(&g0);

  fd = fopen(in_file, "r");
  if (fd == NULL)
    fatal_error("generate_from_description, open failed");
  fscanf(fd, "%d", &g0.n);
  while (fscanf(fd, "%d %d", &j, &k) == 2)
    poset_add_pair(&g0, j, k);
  fclose(fd);

  cpcp(g0);
  count = 0;
  t_max = 0;
  c = snprintf(buffer0, sizeof(buffer0), CPC_PRINT_FORMAT, g0.t);
  for (j = 0; j < N; j++) {
    for (k = 0; k < N; k++) {
      if (j == k)
        printf(" %*s", c, "-");
      else if (in(g0.r[0][j], k))
        printf(" %*s", c, buffer0);
      else if (in(g0.r[1][j], k))
        printf(" %*s", c, "0");
      else {
        g1 = g0;
        poset_add_pair(&g1, j, k);
        cpcp(g1);
        snprintf(buffer1, sizeof(buffer1), CPC_PRINT_FORMAT, g1.t);
        printf(" %*s", c, buffer1);
        if (2 * g1.t <= g0.t) {
          if (g1.t > t_max) {
            t_max = g1.t;
            x[0] = j;
            y[0] = k;
            count = 1;
          }
          else if (g1.t == t_max && count < sizeof(x) / sizeof(x[0])) {
            x[count] = j;
            y[count] = k;
            count++;
          }
        }
      }
    }
    printf("\n");
  }
  printf("b(P) = " CPC_PRINT_FORMAT " / " CPC_PRINT_FORMAT " = %.15Lf,"
         " for (x, y) = ",
         t_max, g0.t, (long double)t_max / (long double)g0.t);
  for (c = 0; c < count - 1; c++)
    printf("(%d, %d), ", x[c], y[c]);
  printf("(%d, %d)\n", x[count - 1], y[count - 1]);

  fh.cut_off = 1;
  if (c_max >= 0)
    for (c = 0; c < c_max; c++)
      fh.cut_off *= 2;
  else
    while (g0.t > fh.cut_off)
      fh.cut_off *= 2;
  if (g0.t > fh.cut_off)
    printf("Cannot sort using %d comparisons.\n", c_max);
  else
    fh.posets_count = 1;

  if (file[OUTF] >= 0) {
    if (write(file[OUTF], &fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("generate_from_description, write file header failed");
    if (fh.posets_count)
      if (write(file[OUTF], &g0, sizeof(POSET)) != sizeof(POSET))
        fatal_error("generate_from_description, write poset failed");
  printf("step %d  input posets %ld  output posets %ld\n",
         fh.step, fh.analysed_count, fh.posets_count);
  }
  return fh.posets_count == 0;
}

int join_posets(void) {
  FILE_HEADER in_fh, out_fh;
  long        l;

  empty_storage(EMPTY_SORTABLE_STACK);
  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("join_posets, read input file header failed");
  if (read(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("join_posets, read output file header failed");
  if (in_fh.version != FILE_VERSION ||
      out_fh.version != FILE_VERSION ||
      in_fh.step != out_fh.step ||
      in_fh.cut_off != out_fh.cut_off)
    fatal_error("join_posets, missing headers");
  reset_read_buffer(OUTF);
  for (out_fh.analysed_count = 0;
       out_fh.analysed_count < out_fh.posets_count;
       out_fh.analysed_count++) {
    #ifdef _PRINT_PROGRESS_
    printf("%9ld %9ld\r", out_fh.analysed_count, (long)sortable_count);
    #endif
    sortable_insert(get_next_poset(NULL));
  }
  reset_read_buffer(INF);
  for (out_fh.analysed_count = out_fh.posets_count;
       out_fh.analysed_count < in_fh.posets_count + out_fh.posets_count;
       out_fh.analysed_count++) {
    #ifdef _PRINT_PROGRESS_
    printf("%9ld %9ld\r", out_fh.analysed_count, (long)sortable_count);
    #endif
    sortable_insert(get_next_poset(NULL));
  }
  #ifdef _PRINT_PROGRESS_
  printf("%9ld %9ld\r", out_fh.analysed_count, (long)sortable_count);
  #endif
  if (lseek(file[OUTF], 0, SEEK_SET) < 0)
    fatal_error("join_posets, lseek(0) failed");
  out_fh.posets_count = sortable_count;
  out_fh.time += in_fh.time;
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("join_posets, write file header failed");
  l = out_fh.posets_count * sizeof(POSET);
  if (write(file[OUTF], sortable_stack, l) != l)
    fatal_error("join_posets, write failed");
  printf("step %d  input posets %ld  output posets %ld\n",
         out_fh.step, out_fh.analysed_count, out_fh.posets_count);
  return out_fh.posets_count == 0;
}

#if THREADS_NUMBER == 1

/* Idea of this functions comes from Gerold Jaeger. */
/* Writing to the output file is not buffered. */
int remove_congruent_posets() {
  FILE_HEADER in_fh_1, in_fh_2, out_fh;
  POSET       *p_g;
  long        l;

  empty_storage(EMPTY_POSETS_STACK);
  if (read(file[IN1F], &in_fh_1, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("remove_congruent_posets: "
                "read the first input file header failed");
  if (read(file[IN2F], &in_fh_2, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("remove_congruent_posets: "
                "read the second input file header failed");
  if (in_fh_1.version != FILE_VERSION ||
      in_fh_2.version != FILE_VERSION /* ||
      in_fh_1.step != in_fh_2.step ||
      in_fh_1.cut_off != in_fh_2.cut_off */)
    fatal_error("remove_congruent_posets: missing headers");
  memcpy(&out_fh, &in_fh_2, sizeof(FILE_HEADER));
  out_fh.posets_count = 0; /* We don't know how many posets will remain. */
  out_fh.analysed_count = in_fh_2.posets_count;
  out_fh.time = 0;
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("remove_congruent_posets: "
                "write(1) output file header (1) failed");

  /* loading first file into hash table */
  reset_read_buffer(IN1F);
  for (l = 0; l < in_fh_1.posets_count; l++) {
    #ifdef _PRINT_PROGRESS_
    printf("%9ld %9ld\r", l, (long)new_posets_count);
    #endif
    p_g = get_next_poset(NULL);
    if (search(p_g) != EMPTY_POS)
      fatal_error("remove_congruent_posets: "
                  "congruent poset found in the first input file");
    else
      insert(p_g);
  }
  #ifdef _PRINT_PROGRESS_
  printf("%9ld %9ld\r", in_fh_1.posets_count, (long)new_posets_count);
  #endif

  /* searching second file for congruent posets */
  reset_read_buffer(IN2F);
  for (l = 0; l < in_fh_2.posets_count; l++) {
    #ifdef _PRINT_PROGRESS_
    printf("%9ld %9ld\r", l, out_fh.posets_count);
    #endif
    p_g = get_next_poset(NULL);
    if (search(p_g) == EMPTY_POS) {
      /* congruence not found */
      if (write(file[OUTF], p_g, sizeof(POSET)) != sizeof(POSET))
        fatal_error("remove_congruent_posets: write poset failed");
      out_fh.posets_count++;
    }
  }
  #ifdef _PRINT_PROGRESS_
  printf("%9ld %9ld\r", in_fh_2.posets_count, out_fh.posets_count);
  #endif

  if (lseek(file[OUTF], 0, SEEK_SET) < 0)
    fatal_error("remove_congruent_poosets: call to lseek(0) failed");
  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("remove_congruent_posets: can't write output file header (2)");
  l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
  if (lseek(file[OUTF], l, SEEK_SET) != l)
    fatal_error("remove_congruent_posets: call to lseek(end) failed");
  printf("step %d  input posets %ld  output posets %ld\n",
         out_fh.step, in_fh_2.posets_count, out_fh.posets_count);
  return out_fh.posets_count == 0;
}

int select_posets(int c_max) {
  FILE_HEADER in_fh, out_fh;
  POSET       *p_g;
  long        l;
  int         i;

  empty_storage(EMPTY_POSETS_STACK);

  if (read(file[INF], &in_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("select_posets, read file header failed");
  if (in_fh.version != FILE_VERSION)
    fatal_error("select_posets, missing input file version %d",
                in_fh.version);

  memcpy(&out_fh, &in_fh, sizeof(FILE_HEADER));
  out_fh.step = 0;
  out_fh.analysed_count = 0;
  out_fh.posets_count = 0;
  out_fh.time = 0;
  out_fh.cut_off = 1; /* cut_off = 2^c_max */
  for (i = 0; i < c_max; i++)
    out_fh.cut_off *= 2;

  if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("select_posets, write output file header failed");
  #ifdef _PRINT_PROGRESS_
  printf("%9ld %9ld\r", out_fh.analysed_count, out_fh.posets_count);
  #endif

  /* in_fh.time is reused as analysis start time stamp and
     out_fh.time is elapsed analysis time */
  in_fh.time = time(NULL) - out_fh.time;

  reset_read_buffer(INF);
  while (out_fh.analysed_count < in_fh.posets_count) {
    p_g = get_next_poset(NULL);

    if (p_g->t <= out_fh.cut_off)
      if (search(p_g) == EMPTY_POS)
        insert(p_g);

    out_fh.analysed_count++;
    if (out_fh.analysed_count % UPDATE_INTERVAL == 0 ||
        out_fh.analysed_count == in_fh.posets_count ||
        new_posets_count >= MAX_STACK) {
      l = (new_posets_count - sav_posets_count) * sizeof(POSET);
      if (write(file[OUTF], posets_stack + sav_posets_count, l) != l)
        fatal_error("select_posets save, write posets failed");
      if (lseek(file[OUTF], 0, SEEK_SET) < 0)
        fatal_error("select_posets save, lseek(0) failed");
      out_fh.posets_count += new_posets_count - sav_posets_count;
      out_fh.time = time(NULL) - in_fh.time;
      if (write(file[OUTF], &out_fh, sizeof(FILE_HEADER)) !=
          sizeof(FILE_HEADER))
        fatal_error("select_posets save, write file header failed");
      l = sizeof(FILE_HEADER) + out_fh.posets_count * sizeof(POSET);
      if (lseek(file[OUTF], l, SEEK_SET) != l)
        fatal_error("select_posets save, lseek(end) failed");
      if (new_posets_count >= MAX_STACK)
        new_posets_count = 0;
      sav_posets_count = new_posets_count;
    }
    #ifdef _PRINT_PROGRESS_
    printf("%9ld %9ld\r", out_fh.analysed_count,
           out_fh.posets_count + new_posets_count - sav_posets_count);
    #endif
  }

  printf("step %d  input posets %ld  output posets %ld  "
         "computation time %ld\n",
         out_fh.step, out_fh.analysed_count,
         out_fh.posets_count, out_fh.time);
  return out_fh.posets_count == 0;
}

#endif /* THREADS_NUMBER == 1 */

/*
 * The Gold Partition Conjecture routines
 */

#if THREADS_NUMBER == 1

static void check_and_insert(POSET *p, int check_type) {
  int j;
  SET A, B, V;

  switch (check_type) {
    case 0: /* not a linear sum                            */
    case 3: /* not a linear sum, width at most 3           */
    case 4: /* not a linear sum, ladders with broken rungs */
      for (B = 1; B <= FULL_SET / 2; B++) {
        V = B;
        A = FULL_SET;
        while (V) {
          j = min_element(V);
          A &= p->r[0][j] | p->r[1][j];
          V ^= pof2(j);
        }
        if ((A ^ B) == FULL_SET)
          return;
      }
      /*
      for (j = 0; j < N; j++)
        if (cardinality(p->r[0][j] | p->r[1][j]) == N - 1)
          return;
      */
      break;
    case 2: /* height 2 */
      for (j = 0; j < N; j++)
        if (p->r[0][j] != EMPTY_SET && p->r[1][j] != EMPTY_SET)
          return;
      break;
  }
  cpcp(*p);
  insert(p);
}

static BOOL is_triple_balanced(POSET p, int x, int y, int z,
                               CPC t_yx, CPC t_zy) {

  if (in(p.r[0][x], y) || in(p.r[0][y], z))
    return FALSE; /* We assume that there is no 1/2-balanced pair. */
  if (in(p.r[1][x], y))
    fatal_error("is_triple_balanced, x > y");
  else
    poset_add_pair(&p, x, y);
  if (in(p.r[1][y], z))
    fatal_error("is_triple_balanced, y > z");
  else
    poset_add_pair(&p, y, z);
  cpcp(p);
  return p.t <= t_yx || p.t <= t_zy;
}

#define UNSLAVED_PAIR        0
#define NOT_2_SLAVED_PAIR    1
#define BALANCED_PAIR        2
#define CYCLIC_TRIPLE        3
#define BALANCED_TRIPLE      4
#define GPC_FAILED           5
#define MAX_GPC_STATISTICS   6

static int gpc_test_0(POSET *p0) {
  POSET p1;
  CPC   t[N][N];
  int   j, k, x, y, z;
  BOOL  balanced_pair, balanced_triple;

  balanced_pair = FALSE;
  for (j = 0; j < N - 1; j++) {
    for (k = j + 1; k < N; k++) {
      if (in(p0->r[0][j], k)) {
        t[j][k] = p0->t;
        t[k][j] = 0;
      }
      else if (in(p0->r[1][j], k)) {
        t[j][k] = 0;
        t[k][j] = p0->t;
      }
      else {
        /*
        if (cardinality(p0->r[0][j] & ~p0->r[0][k]) +
            cardinality(p0->r[1][k] & ~p0->r[1][j]) <= 1 &&
            cardinality(p0->r[0][k] & ~p0->r[0][j]) +
            cardinality(p0->r[1][j] & ~p0->r[1][k]) <= 1)
          return UNSLAVED_PAIR;
        */
        p1 = *p0;
        if (cardinality(p0->r[1][j]) + cardinality(p0->r[0][k]) >=
            cardinality(p0->r[0][j]) + cardinality(p0->r[1][k])) {
          poset_add_pair(&p1, j, k);
          cpcp(p1);
          t[j][k] = p1.t;
          t[k][j] = p0->t - p1.t;
        }
        else {
          poset_add_pair(&p1, k, j);
          cpcp(p1);
          t[k][j] = p1.t;
          t[j][k] = p0->t - p1.t;
        }
        if (t[j][k] >= t[k][j] &&
            cardinality(p0->r[0][j] & ~p0->r[0][k]) +
            cardinality(p0->r[1][k] & ~p0->r[1][j]) <= 1)
          return NOT_2_SLAVED_PAIR;
        if (t[j][k] <= t[k][j] &&
            cardinality(p0->r[0][k] & ~p0->r[0][j]) +
            cardinality(p0->r[1][j] & ~p0->r[1][k]) <= 1)
          return NOT_2_SLAVED_PAIR;
        if (t[j][k] == t[k][j])
          balanced_pair = TRUE;
      }
    }
  }
  if (balanced_pair)
    return BALANCED_PAIR;

  balanced_triple = FALSE;
  for (z = 2; z < N; z++) {
    for (y = 1; y < z; y++) {
      for (x = 0; x < y; x++) {
        /* There is no 1/2-balanced pair. */
        switch ((2 * t[x][y] > p0->t) |
                ((2 * t[y][z] > p0->t) << 1) |
                ((2 * t[x][z] > p0->t) << 2)) {
          case 0: /* x > y > z */
            if (!balanced_triple &&
                is_triple_balanced(*p0, z, y, x, t[y][z], t[x][y]))
              balanced_triple = TRUE;
            break;
          case 1: /* y > x > z */
            if (!balanced_triple &&
                is_triple_balanced(*p0, z, x, y, t[x][z], t[y][x]))
              balanced_triple = TRUE;
            break;
          case 2: /* x > z > y */
            if (!balanced_triple &&
                is_triple_balanced(*p0, y, z, x, t[z][y], t[x][z]))
              balanced_triple = TRUE;
            break;
          case 3: /* x < y, y < z, z < x */
          case 4: /* x > y, y > z, z > x */
            // printf("(%d, %d, %d)\n", x, y, z);
            // poset_print(p0);
            return CYCLIC_TRIPLE;
          case 5: /* y > z > x */
            if (!balanced_triple &&
                is_triple_balanced(*p0, x, z, y, t[z][x], t[y][z]))
              balanced_triple = TRUE;
            break;
          case 6: /* z > x > y */
            if (!balanced_triple &&
                is_triple_balanced(*p0, y, x, z, t[x][y], t[z][x]))
              balanced_triple = TRUE;
            break;
          case 7: /* z > y > x */
            if (!balanced_triple &&
                is_triple_balanced(*p0, x, y, z, t[y][x], t[z][y]))
              balanced_triple = TRUE;
            break;
        }
      }
    }
  }
  if (balanced_triple)
    return BALANCED_TRIPLE;
  poset_print(p0);
  return GPC_FAILED;
}

static long double b_test(POSET *p0) {
  POSET       p1;
  CPC         t_jk, t_kj;
  int         j, k;
  long double b, b_max;

  b_max = 0.0;
  for (j = 0; j < N - 1; j++) {
    for (k = j + 1; k < N; k++) {
      if (!in(p0->r[0][j], k) && !in(p0->r[1][j], k)) {
        p1 = *p0;
        if (cardinality(p0->r[1][j]) + cardinality(p0->r[0][k]) >=
            cardinality(p0->r[0][j]) + cardinality(p0->r[1][k])) {
          poset_add_pair(&p1, j, k);
          cpcp(p1);
          t_jk = p1.t;
          t_kj = p0->t - p1.t;
        }
        else {
          poset_add_pair(&p1, k, j);
          cpcp(p1);
          t_kj = p1.t;
          t_jk = p0->t - p1.t;
        }
        if (t_jk == t_kj) {
          b_max = 0.5;
          return b_max;
        }
        else if (t_jk > t_kj) {
          b = (long double)t_kj / (long double)p0->t;
          if (b > b_max)
            b_max = b;
        }
        else {
          b = (long double)t_jk / (long double)p0->t;
          if (b > b_max)
            b_max = b;
        }
      }
    }
  }
  return b_max;
}

int gpc_iterative_test(int test_type, int check_type) {
  POSET       *p0, p1;
  long        timer;
  INDEX       l, from, to;
  int         j, k, step, count;
  long        statistics[MAX_GPC_STATISTICS];
  POSET       p2[10];
  long double b_min, b_max;
  #ifdef _WIDTH_2_
  SET         rungs_set;
  #else
  int         i;
  #endif

  timer = time(NULL);
  empty_storage(EMPTY_POSETS_STACK);
  step = 0;
  to = 0;

  #ifdef _WIDTH_2_
  switch (check_type) {
    case 0: /* not a linear sum, width 2 */
      poset_empty(&p1);
      for (j = 1; j < N; j++)
        if (j != M)
          poset_add_pair(&p1, j - 1, j);
      p1.n = N;
      cpcp(p1);
      if (search(&p1) == EMPTY_POS)
        insert(&p1);
      break;
    case 4: /* not a linear sum, ladders with broken rungs */
      #if M < N - M - 2
        #define RUNGS_COUNT (2 * M - 1)
      #else
        #define RUNGS_COUNT (N - 3)
      #endif
      for (rungs_set = EMPTY_SET; rungs_set < pof2(RUNGS_COUNT); rungs_set++) {
        poset_empty(&p1);
        for (j = 1; j < N; j++)
          if (j != M)
            poset_add_pair(&p1, j - 1, j);
        for (j = 0; j < RUNGS_COUNT; j++) {
          if (in(rungs_set, j)) {
            if (j < M - 1)
              poset_add_pair(&p1, j + M, j + 1);
            else
              poset_add_pair(&p1, j - M + 1, j + 3);
          }
        }
        p1.n = N;
        if (search(&p1) == EMPTY_POS) {
          cpcp(p1);
          insert(&p1);
          if (new_posets_count > MAX_STACK) {
            printf("gpc_test, posets stack overflow\n");
            break;
          }
        }
      }
      to = new_posets_count;
      #undef RUNGS_COUNT
      break;
    default:
      fatal_error("gpc_test, not implemented check_type %d", check_type);
  }
  #else /* _WIDTH_2_ not defined */
  switch (check_type) {
    case 0: /* not a linear sum */
    case 2: /* height 2         */
      poset_empty(&p1);
      p1.n = N;
      cpcp(p1);
      if (search(&p1) == EMPTY_POS)
        insert(&p1);
      break;
    case 3: /* not a linear sum, width at most 3 */
      for (j = 1; j < N; j++) {
        for (k = j; k < N; k++) {
          poset_empty(&p1);
          for (i = 1; i < N; i++)
            if (i != j && i != k)
              poset_add_pair(&p1, i - 1, i);
          p1.n = N;
          if (search(&p1) == EMPTY_POS) {
            cpcp(p1);
            insert(&p1);
          }
        }
      }
      break;
    default:
      fatal_error("gpc_test, not implemented check_type %d", check_type);
  }
  #endif /* _WIDTH_2_ */

  printf("Generating %d-element posets\n"
         "step %d  input posets %ld  output posets %ld\n",
         N, step, (long)to, (long)new_posets_count);
  while (to < new_posets_count && new_posets_count <= MAX_STACK) {
    step++;
    from = to;
    to = new_posets_count;
    for (l = from; l < to; l++) {
      p0 = posets_stack + l;
      for (j = 0; j < N; j++) {
        for (k = 0; k < N; k++) {
          if (((p0->r[0][j] | p0->r[1][j] | pof2(j)) & pof2(k)) == EMPTY_SET) {
            p1 = *p0;
            poset_add_pair(&p1, j, k);
            if (search(&p1) == EMPTY_POS)
              check_and_insert(&p1, check_type);
          }
        }
      }
      if (new_posets_count > MAX_STACK) {
        printf("gpc_test, posets stack overflow\n");
        break;
      }
    }
    printf("step %d  input posets %ld  output posets %ld\n",
           step, (long)to, (long)new_posets_count);
  }
  printf("computation time  %9ld\n\n", time(NULL) - timer);

  if (test_type & 1) { /* GPC */
    for (j = 0; j < MAX_GPC_STATISTICS; j++)
      statistics[j] = 0;
    for (l = 0; l < new_posets_count; l++)
      statistics[gpc_test_0(posets_stack + l)]++;
    printf("The Gold Partition Conjecture\n"
           "analysed posets   %9ld\n"
           "unslaved pair     %9ld\n"
           "at most 1 slave   %9ld\n"
           "1/2-balanced pair %9ld\n"
           "cyclic triple     %9ld\n"
           "balanced triple   %9ld\n"
           "failed            %9ld\n"
           "computation time  %9ld\n\n",
           (long)new_posets_count,
           statistics[UNSLAVED_PAIR],
           statistics[NOT_2_SLAVED_PAIR],
           statistics[BALANCED_PAIR],
           statistics[CYCLIC_TRIPLE],
           statistics[BALANCED_TRIPLE],
           statistics[GPC_FAILED],
           time(NULL) - timer);
  }
  if (test_type & 2) { /* the worst balanced poset */
    b_min = 1.0;
    count = 0;
    for (l = 0; l < new_posets_count; l++) {
      b_max = b_test(posets_stack + l);
      if (b_max == b_min) {
        if (count < sizeof(p2) / sizeof(p2[0])) {
          p2[count] = *(posets_stack + l);
          count++;
        }
      }
      else if (b_max < b_min) {
        count = 1;
        b_min = b_max;
        p2[0] = *(posets_stack + l);
      }
    }
    printf("The Worst Balanced Poset\n");
    for (j = 0; j < count; j++)
      poset_print(p2 + j);
    printf("b(P)=%.15Lf\n"
           "analysed posets   %9ld\n"
           "computation time  %9ld\n\n",
           b_min,
           (long)new_posets_count,
           time(NULL) - timer);
  }
  if (test_type & 4) { /* B-set */
    for (l = 0; l < new_posets_count; l++)
      printf("%.15Lf\n", b_test(posets_stack + l));
    printf("B-set\n"
           "analysed posets   %9ld\n"
           "computation time  %9ld\n\n",
           (long)new_posets_count,
           time(NULL) - timer);
  }
  return 0;
}

#endif /* THREADS_NUMBER == 1 */
